"""
ScrapeMed's Custom Markup Language - MoreHTML (MHTML)
======================================================

Wrapper on basic functions for HTML manipulation.

**Added on top of core html functionality:**
Non-markup significant unescape function, custom MHTML tag encoding and removal.
"""

import re
import html


def unescape_except(s, **kwargs):
    """
    Convert all named and numeric character references in the provided string to
    the corresponding Unicode characters, excluding any provided encodings to be
    ignored.

    :param str s: The input string containing character references.
    :param kwargs: Keyword arguments of the form key=encoding. These encodings
            will be ignored when unescaping.
                    For keys with multiple encodings, use unique keynames.
                   Encodings must be single code strings.
    :type kwargs: dict
    :return: A string with character references unescaped, except for the
        specified encodings to be ignored.
    :rtype: str

    This function uses the rules defined by the HTML 5 standard for both valid
    and invalid character references, and the list of HTML 5 named character
    references defined in html.entities.html5.
    """

    # no need to do anything if there are no html encodings
    if "&" not in s:
        return s

    encoding_dict = {}

    # Translate keys to MHTML placeholder codes
    for key, encoding in kwargs.items():
        placehold_str = generate_mhtml_tag(key)
        encoding_dict[placehold_str] = encoding

    # Convert encodings to MHTML placeholder codes
    for placehold_str, encoding in encoding_dict.items():
        code_to_save = re.compile(re.escape(encoding))
        s = code_to_save.sub(placehold_str, s)

    # Unescape everything else
    s = html.unescape(s)

    # Convert placeheld items back to their original html encodings
    for placehold_str, encoding in encoding_dict.items():
        placehold_r = re.compile(re.escape(placehold_str))
        s = placehold_r.sub(encoding, s)

    return s


def generate_mhtml_tag(string: str) -> str:
    """
    Generates an MHTML tag from the provided string.

    :param str string: The text to be tagged in MHTML format.
    :return: An MHTML tag containing the input string, in format
        `f"[MHTML::{string}]"`
    :rtype: str
    """
    return f"[MHTML::{string}]"


def generate_typed_mhtml_tag(tag_type: str, string: str) -> str:
    """
    Generates a typed MHTML tag from the provided string.

    :param str tag_type: The type of the MHTML tag.
    :param str string: The text to be tagged in MHTML format.
    :return: A typed MHTML tag containing the input string, in format
        `[MHTML::type::string]`.
    :rtype: str
    """
    return f"[MHTML::{tag_type}::{string}]"


def remove_mhtml_tags(text: str) -> str:
    """
    Removes all MHTML tags and typed MHTML tags found in the provided text.

    :param str text: The text from which to remove MHTML tags.
    :return: The text with MHTML tags removed.
    :rtype: str
    """
    # match MHTML tags
    # group1 = tag type for typed MHTML tags
    # group2 = tag value for typed MHTML tags
    # group3 = tag for non-typed MHTML tags
    mhtml_pattern = r"\[MHTML::([^:\[\]]+)::([^:\[\]]+)\]" r"|\[MHTML::([^:\[\]]+)\]"
    mhtml_r = re.compile(mhtml_pattern)
    # remove MHTML tags and return result
    return mhtml_r.sub("", text)



"""
ScrapeMed's ``_validate`` Module
=================================

Validation module for determining whether XML conforms to a format
supported by the scrapemed package (NLM Articleset 2.0 DTD).

**Custom Exception**:
    - ``noDTDFoundError``: Raised when no DTD specification can be found in the
        downloaded XML.
"""

import re
import lxml.etree as ET
import os
from io import StringIO
from scrapemed.utils import cleanerdoc

SUPPORTED_DTD_URLS = [
    "https://dtd.nlm.nih.gov/ncbi/pmc/articleset/nlm-articleset-2.0.dtd"
]
# Regex DTD URL Patterns
DTD_URL_PATTERN = re.compile(r'"(https?://\S+)"')
END_OF_URL_PATTERN = re.compile(r"[^/]+$")


class noDTDFoundError(Exception):
    """
    Raised when no DTD specification can be found in a downloaded XML,
    preventing validation.
    """

    pass


# ---------------------------DATA VALIDATION-------------------------------
def validate_xml(xml: ET.ElementTree) -> bool:
    """
    Validate an XML ElementTree against a supported Document Type Definition
    (DTD).

    This function validates the provided XML ElementTree against a supported
    DTD (Document Type Definition). The supported DTDs are defined by the files
    in the 'scrapemed/data/DTDs' directory. Currently only NLM Articleset 2.0
    (The DTD used by PubMed Central) is supported.

    :param ET.ElementTree xml: An XML ElementTree to be validated.

    :return: True if the XML is validated successfully against a supported DTD,
        False otherwise.
    :rtype: bool

    :raises noDTDFoundError: If no DTD is specified for validation in the
        XML doctype.
    """
    # Find DTD and confirm its supported
    match = DTD_URL_PATTERN.search(xml.docinfo.doctype)
    url = None
    if match:
        url = match.group(1)
        assert url in SUPPORTED_DTD_URLS
    else:
        raise noDTDFoundError(
            cleanerdoc(
                """A DTD must be specified for validation. Set
                       validate=false if you want to proceeed without
                       validation."""
            )
        )

    match = END_OF_URL_PATTERN.search(url)
    dtd_filename = match.group(0)
    dtd_filepath = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "data", "DTDs", dtd_filename
    )

    dtd_doc = None
    with open(dtd_filepath, "r") as f:
        dtd_doc = f.read()
    if dtd_doc is None:
        raise noDTDFoundError(
            cleanerdoc(
                """DTD not found in scrapemed package. Ensure you are using the
                latest package version."""
            )
        )

    dtd = ET.DTD(StringIO(dtd_doc))

    return dtd.validate(xml)


# -------------------------END DATA VALIDATION-------------------------------



"""
ScrapeMed's Paper Module
============================

The scrapemed `paper` module is intended as the primary point of contact for
scrapemed end users.

Paper objects are defined here, as well end-user functionality for scraping
data from PubMed Central without stressing about the details.

..warnings::
    - :class:`emptyTextWarning` - Warned when trying to perform a text
        operation on a Paper which has no text.
    - :class:`pubmedHTTPError` - Warned when unable to retrieve a PMC XML
        repeatedly. Can occasionally happen with PMC due to high traffic.
        Also may be caused by broken XML formatting.
"""

import scrapemed._parse as parse
import lxml.etree as ET
import pandas as pd
import datetime
import chromadb
from langchain.text_splitter import CharacterTextSplitter
from typing import Union, Dict
from difflib import SequenceMatcher
import uuid
import re
import warnings
from urllib.error import HTTPError
import time


class emptyTextWarning(Warning):
    """
    Warned when trying to perform a text operation on a Paper which has no text.
    """

    pass


class pubmedHTTPError(Warning):
    """
    Warned when unable to retrieve a PMC XML repeatedly. Can occasionally
    happen with PMC due to high traffic. Also may be caused by broken XML
    formatting.
    """

    pass


# --------------------PAPER OBJECT SCHEMA-------------------------------------
class Paper:
    """
    Class for storing paper data downloaded from PMC.

    This class provides methods for initializing papers via PMCID and directly
    from XML, paper chunking and vectorization, conversion to relational format
    (pandas Series), printing methods, and equality checking.

    Class data members include all of the data defined via the method
    :meth:`~Paper.info`.

    :raises pubmedHTTPError: Raised if there are HTTP errors when
        retrieving data from PMC.
    :raises emptyTextWarning: Raised if an attempt is made to vectorize a paper
        with no text.

    :Example:

    To initialize a Paper object with paper_dict:

    >>> paper = Paper(paper_dict)
    """

    __tablename__ = "Papers"

    def __init__(self, paper_dict: dict) -> None:
        """
        Initialize a Paper object with paper information parsed from a PMC download.

        :param dict paper_dict: A dictionary containing paper information, typically
            obtained from the parse.generate_paper_dict method.
        """
        if not paper_dict:
            self.has_data = False
            return None
        else:
            self.has_data = True

        # capture current time as time of last update. Note that this date
        # may not be synced with PMC paper updates if using
        # initialization via Paper.from_xml. Use Paper.from_pmc to update
        # papers directly via PMC
        current_datetime = datetime.datetime.now()
        current_year = current_datetime.year
        current_month = current_datetime.month
        current_day = current_datetime.day
        self.last_updated = (current_month, current_day, current_year)

        # read in the Paper data from the parsed paper_dict
        self.pmcid = paper_dict["PMCID"]
        self.title = paper_dict["Title"]
        self.authors = paper_dict["Authors"]
        self.non_author_contributors = paper_dict["Non-Author Contributors"]
        self.abstract = paper_dict["Abstract"]
        self.body = paper_dict["Body"]
        self.journal_id = paper_dict["Journal ID"]
        self.journal_title = paper_dict["Journal Title"]
        self.issn = paper_dict["ISSN"]
        self.publisher_name = paper_dict["Publisher Name"]
        self.publisher_location = paper_dict["Publisher Location"]
        self.article_id = paper_dict["Article ID"]
        self.article_types = paper_dict["Article Types"]
        self.article_categories = paper_dict["Article Categories"]
        self.published_date = paper_dict["Published Date"]
        self.volume = paper_dict["Volume"]
        self.issue = paper_dict["Issue"]
        self.fpage = paper_dict["First Page"]
        self.lpage = paper_dict["Last Page"]
        self.permissions = paper_dict["Permissions"]
        if self.permissions:
            self.copyright = self.permissions["Copyright Statement"]
            self.license = self.permissions["License Type"]
        else:
            self.copyright = None
            self.license = None
        self.funding = paper_dict["Funding"]
        self.footnote = paper_dict["Footnote"]
        self.acknowledgements = paper_dict["Acknowledgements"]
        self.notes = paper_dict["Notes"]
        self.custom_meta = paper_dict["Custom Meta"]
        self.ref_map = paper_dict["Ref Map"]
        self._ref_map_with_tags = paper_dict["Ref Map With Tags"]
        self.citations = paper_dict["Citations"]
        self.tables = paper_dict["Tables"]
        self.figures = paper_dict["Figures"]

        self.data_dict = parse.define_data_dict()

        self.vector_collection = None

        return None

    @classmethod
    def from_pmc(
        cls,
        pmcid: int,
        email: str,
        download: bool = False,
        validate: bool = True,
        verbose: bool = False,
        suppress_warnings: bool = False,
        suppress_errors: bool = False,
    ):
        """
        Generate a Paper from a PMCID with optional parameters.

        :param int pmcid: Unique PMCID for the article to parse.
        :param str email: Provide your email address for authentication with PMC.
        :param bool download: Whether or not to download the XML retrieved from PMC.
        :param bool validate: Whether or not to validate the XML from PMC against NLM
            articleset 2.0 DTD (HIGHLY RECOMMENDED).
        :param bool verbose: Whether or not to have verbose output for testing.
        :param bool suppress_warnings: Whether to suppress warnings while parsing XML.
            Note: Warnings are frequent, because of the variable nature of PMC
            XML data. Recommended to suppress when parsing many XMLs at once.
        :param bool suppress_errors: Return None on failed XML parsing, instead of
            raising an error.

        :return: A Paper object initialized via the passed PMCID and
            optional parameters.
        :rtype: Paper
        """
        NUM_TRIES = 3
        paper_dict = None
        for i in range(NUM_TRIES):
            try:
                paper_dict = parse.paper_dict_from_pmc(
                    pmcid=pmcid,
                    email=email,
                    download=download,
                    validate=validate,
                    verbose=verbose,
                    suppress_warnings=suppress_warnings,
                    suppress_errors=suppress_errors,
                )
                break
            except HTTPError:
                time.sleep(5)
        if not paper_dict:
            warnings.warn(
                (
                    f"Unable to retrieve PMCID {pmcid} from PMC. May be due to "
                    "HTTP traffic or broken XML formatting, try again later if "
                    "the former."
                ),
                pubmedHTTPError,
            )
            return None
        return cls(paper_dict)

    @classmethod
    def from_xml(
        cls,
        pmcid: int,
        root: ET.Element,
        verbose: bool = False,
        suppress_warnings: bool = False,
        suppress_errors: bool = False,
    ):
        """
        Generate a Paper straight from PMC XML.

        :param int pmcid: PMCID for the XML. THis is required intentionally,
            to ensure trustworthy unique indexing of PMC XMLs.
        :param ET.Element root: Root element of the PMC XML tree.
        :param bool verbose: Report verbose output or not. Intended for testing.
        :param bool suppress_warnings: Suppress warnings while
            parsing XML or not.
            Note: Warnings are frequent, because of the variable nature of
            PMC XML data.
            Recommended to suppress when parsing many XMLs at once.
        :param bool suppress_errors: Return None on failed XML parsing,
            instead of raising an error.
            Recommended to suppress when parsing many XMLs at once, unless
            failure is not an option.

        :returns: A Paper object initialized via the passed XML.
        :rtype: Paper
        """
        paper_dict = parse.generate_paper_dict(
            pmcid,
            root,
            verbose=verbose,
            suppress_warnings=suppress_warnings,
            suppress_errors=suppress_errors,
        )
        return cls(paper_dict)

    def info(self) -> Dict[str, str]:
        """
        Return the data definition dictionary.

        :return: A dictionary containing paper information.
        :rtype: dict[str, str]
        """
        return self.data_dict

    def print_abstract(self) -> str:
        """
        Print and return a string representation of the abstract.

        :return: A string containing the abstract text.
        :rtype: str
        """
        s = self.abstract_as_str()
        print(s)
        return s

    def abstract_as_str(self) -> str:
        """
        Return a string representation of the abstract of a paper.

        This method retrieves the abstract text without MHTML data references.

        :return: A string containing the abstract text.
        :rtype: str
        """
        s = ""
        if self.abstract:
            for sec in self.abstract:
                s += "\n"
                s += str(sec)
        return s

    def print_body(self) -> str:
        """
        Print and return a string representation of the body of a paper.

        This method retrieves the body text without MHTML data references.

        :return: A string containing the body text.
        :rtype: str
        """
        s = self.body_as_str()
        print(s)
        return s

    def body_as_str(self) -> str:
        """
        Return a string representation of the body of a paper.

        :return: A string containing the body text.
        :rtype: str
        """
        s = ""
        if self.body:
            for sec in self.body:
                s += "\n"
                s += str(sec)
        return s

    def __bool__(self):
        """
        Determine the truth value of a Paper object based on successful
        initialization.

        :return: True if the Paper object was successfully initialized with
            data, False otherwise.
        :rtype: bool
        """
        return self.has_data

    def full_text(self, print_text: bool = False):
        """
        Return the full abstract and/or body text of this Paper as a string.

        Optionally, you can choose to print the text.

        :param bool print_text: If True, print the text; if False, return it
            as a string.

        :return: A string containing the full text of the abstract and/or body.
        :rtype: str
        """
        s = ""
        if self.abstract:
            s += "Abstract: \n"
            s += self.abstract_as_str()
        if self.body:
            s += "Body: \n"
            s += self.body_as_str()

        if print_text:
            print(s)
        return s

    def __str__(self):
        """
        Return a string representation of the Paper object.

        :return: A string containing the PMCID, title, abstract, and body text
            of the paper.
        :rtype: str
        """
        s = ""
        s += f"\nPMCID: {self.pmcid}\n"
        s += f"Title: {self.title}\n"
        # Append all text from abstract PaperSections
        s += "\nAbstract:\n"
        if self.abstract:
            for sec in self.abstract:
                s += str(sec)
        # Append all text from body PaperSections
        s += "\nBody:\n"
        if self.body:
            for sec in self.body:
                s += str(sec)
        return s

    def __eq__(self, other):
        """
        Check if two Paper objects are equal.

        Two Paper objects are considered equal if they share the same PMCID and have
        the same date of last update. Papers with the same content but downloaded or
        parsed on different dates are not considered equal.

        To compare Paper objects based solely on their PMCID, use
        `Paper1.pmcid == Paper2.pmcid`.

        Note that articles that are not open access on PMC may not have a PMCID, and a
        unique comparison method will be needed for these cases. However, most papers
        downloaded via ScrapeMed should have a PMCID.

        :param other: The other Paper object to compare.
        :type other: Paper
        :return: True if the two Paper objects are equal, False otherwise.
        :rtype: bool
        """
        if not self:
            return False
        return self.pmcid == other.pmcid and self.last_updated == other.last_updated

    def to_relational(self) -> pd.Series:
        """
        Generate a pandas Series representation of the paper.

        This method creates a pandas Series containing a relational representation of
        the paper's data. Some data may be lost in this process, but most useful text
        data and metadata will be retained in a structured form.

        :return: A pandas Series representing the paper's data.
        :rtype: pd.Series
        """

        data = {
            "PMCID": self.pmcid,
            "Last_Updated": self.last_updated,
            "Title": self.title,
            "Authors": self._extract_names(self.authors)
            if isinstance(self.authors, pd.DataFrame)
            else None,
            "Non_Author_Contributors": self._extract_names(self.non_author_contributors)
            if isinstance(self.non_author_contributors, pd.DataFrame)
            else None,
            "Abstract": self.abstract_as_str(),
            "Body": self.body_as_str(),
            "Journal_ID": self.journal_id,
            "Journal_Title": self.journal_title,
            "ISSN": self.issn,
            "Publisher_Name": self.publisher_name,
            "Publisher_Location": self.publisher_location,
            "Article_ID": self.article_id,
            "Article_Types": self.article_types,
            "Article_Categories": self.article_categories,
            "Published_Date": self._serialize_dict(self.published_date)
            if isinstance(self.published_date, dict)
            else None,
            "Volume": self.volume,
            "Issue": self.issue,
            "First_Page": self.fpage,
            "Last_Page": self.lpage,
            "Copyright": self.copyright,
            "License": self.license,
            "Funding": self.funding,
            "Footnote": self.footnote,
            "Acknowledgements": self.acknowledgements,
            "Notes": self.notes,
            "Custom_Meta": self.custom_meta,
            "Ref_Map": self.ref_map,
            "Citations": [
                self._serialize_dict(c) for c in self.citations if isinstance(c, dict)
            ],
            "Tables": [
                self._serialize_df(t)
                for t in self.tables
                if isinstance(t, (pd.io.formats.style.Styler, pd.DataFrame))
            ],
            "Figures": self.figures,
        }
        return pd.Series(data)

    # ---------------Helper functions for to_relational---------------------
    def _extract_names(self, df):
        """
        Extract and format names from a DataFrame.

        :param df: The DataFrame containing name data.
        :type df: pd.DataFrame
        :return: A list of formatted names.
        :rtype: List[str]
        """
        return df.apply(
            lambda row: f"{row['First_Name']} {row['Last_Name']}", axis=1
        ).tolist()

    def _serialize_dict(self, data_dict):
        """
        Serialize a dictionary into a string.

        :param data_dict: The dictionary to serialize.
        :type data_dict: dict
        :return: A string representation of the serialized dictionary.
        :rtype: str
        """
        return "; ".join([f"{key}: {value}" for key, value in data_dict.items()])

    def _serialize_df(self, df):
        """
        Serialize a DataFrame into an HTML string.

        :param df: The DataFrame to serialize.
        :type df: pd.DataFrame
        :return: An HTML representation of the serialized DataFrame.
        :rtype: str
        """

        return df.to_html()

    # ---------------End Helper functions for to_relational--------------------

    def vectorize(
        self, chunk_size: int = 100, chunk_overlap: int = 20, refresh: bool = False
    ):
        """
        Generate an in-memory vector database representation of the paper.

        This method generates an in-memory vector database representation of the
        paper, stored in `paper.vector_collection`. It focuses on vectorizing the
        abstract and body text.

        :param int chunk_size: An approximate chunk size to split the paper into
            (measured in characters).
        :param int chunk_overlap: An approximate desired chunk overlap
            (measured in characters).
        :param bool refresh: Whether or not to clear and re-vectorize the paper
            with new settings.

        :return: None
        """
        if not refresh and self.vector_collection:
            print(
                (
                    "Paper already vectorized! To re-vectorize with new "
                    "settings, pass refresh=True."
                )
            )
            return None

        print("Vectorizing Paper (This may take a little while)...")
        if len(self.full_text()) == 0:
            warnings.warn(
                "Attempted to vectorize a Paper with no text. Aborting.",
                emptyTextWarning,
            )
            return None

        # Set up an in-memory chromadb collection for this paper
        client = chromadb.Client()
        try:
            self.vector_collection = client.get_or_create_collection(
                f"Paper-PMCID-{self.pmcid}"
            )
        except AttributeError:
            self.vector_collection = client.get_or_create_collection(
                f"Paper-Random-UUID-{uuid.uuid4()}"
            )

        # setup chunk model
        chunk_model = CharacterTextSplitter(
            separator="\\n\\n|\\n|\\.|\\s",
            is_separator_regex=True,
            chunk_size=chunk_size,
            chunk_overlap=chunk_overlap,
            length_function=len,
            keep_separator=True,
        )

        # chunk the text, add metadata for the PMCID each chunk
        # originates from, add unique chunk ids
        p_chunks = chunk_model.split_text(self.full_text())
        p_metadatas = [{"pmcid": self.pmcid}] * len(p_chunks)
        try:
            pmcid = self.pmcid
        except AttributeError:
            pmcid = uuid.uuid4()
        p_ids = [self._generate_chunk_id(pmcid, i) for i in range(len(p_chunks))]

        # upload the chunked texts into the vector collection
        self.vector_collection.add(documents=p_chunks, metadatas=p_metadatas, ids=p_ids)

        print(
            (
                "Done Vectorizing Paper! Natural language query with "
                "Paper.query() now available."
            )
        )
        return None

    # -----------------helper funcs for self.vectorize-----------------
    def _generate_chunk_id(self, pmcid: str, index: Union[int, str]):
        """
        Generate an ID for a PMC text chunk using the PMCID and the chunk's index.

        The chunk indices should be unique. It is recommended to use indexes from
        the result of the chunk model.

        :param str pmcid: The PMCID of the paper.
        :param Union[int, str] index: The index of the chunk.
        :return: A unique chunk ID.
        :rtype: str
        """
        return f"pmcid-{pmcid}-chunk-{str(index)}"

    def _get_chunk_index_from_chunk_id(self, chunk_id: str) -> str:
        """
        Given a PMCID Chunk ID in the format generated by `_generate_chunk_id`,
        extract the index of the chunk.

        :param str chunk_id: The chunk ID.
        :return: The index of the chunk.
        :rtype: str
        """
        pattern = re.compile(r"chunk-(\d+)")  # Compile the regex pattern
        match = pattern.search(chunk_id)
        index = None
        if match:
            index = match.group(1)
        return index

    def _get_pmcid_from_chunk_id(self, chunk_id: str) -> str:
        """
        Given a PMCID Chunk ID in the format generated by `_generate_chunk_id`,
        extract the PMCID of the chunk.

        :param str chunk_id: The chunk ID.
        :return: The PMCID of the chunk.
        :rtype: str
        """
        pattern = re.compile(r"pmcid-(\d+)")  # Compile the regex pattern
        match = pattern.search(chunk_id)
        pmcid = None
        if match:
            pmcid = match.group(1)
        return pmcid

    # -----------------end helper funcs for self.vectorize-----------------

    def query(
        self, query: str, n_results: int = 1, n_before: int = 2, n_after: int = 2
    ) -> Dict[str, str]:
        """
        Query the paper with natural language questions.

        :param str query: The natural language question/query.
        :param int n_results: The number of most semantically similar paper
            sections to retrieve.
        :param int n_before: The number of chunks before the match to include
            in the combined output.
        :param int n_after: The number of chunks after the match to include in
            the combined output.

        :return: A dictionary with keys representing the most semantically
            similar result chunk(s) and values representing the paper text(s)
            around the most semantically similar result chunk(s).
            The text length is determined by the chunk size used in
            `self.vectorize()` and the params `n_before` and `n_after`.
        :rtype: dict[str, str]
        """

        result = self.expanded_query(
            query=query, n_results=n_results, n_before=n_before, n_after=n_after
        )

        return result

    # -----------------helper funcs for self.query----------------------
    def expanded_query(
        self, query: str, n_results: int = 1, n_before: int = 2, n_after: int = 2
    ) -> Dict[str, str]:
        """
        Query the paper with an expanded natural language question/query.

        This method matches a natural language query with the vectorized Paper.
        It retrieves and expands the text sections around the most semantically
        similar result chunk(s).

        :param str query: The natural language query.
        :param int n_results: The number of most semantically similar paper
            sections to retrieve.
        :param int n_before: The number of chunks before the match to include
            in the combined output.
        :param int n_after: The number of chunks after the match to include
            in the combined output.

        :return: A dictionary with keys representing the most semantically
            similar result chunk(s) and values representing the expanded paper
            text(s) around the result chunk(s).
        :rtype: dict[str, str]
        """
        # if the paper has not already been vectorized, vectorize
        if not self.vector_collection:
            self.vectorize()
        # if vectorization fails, abort
        if not self.vector_collection:
            return None

        result = self.vector_collection.query(
            query_texts=[query], include=["documents"], n_results=n_results
        )

        expanded_results = {}
        for id in result["ids"][0]:
            chunk_index = self._get_chunk_index_from_chunk_id(id)
            pmcid = self._get_pmcid_from_chunk_id(id)
            # get the texts before and after the result chunk
            expanded_ids = []
            for i in range(1, n_before + 1):
                expanded_ids.append(
                    self._generate_chunk_id(pmcid, int(chunk_index) - i)
                )
            expanded_ids.append(id)
            for i in range(1, n_after + 1):
                expanded_ids.append(
                    self._generate_chunk_id(pmcid, int(chunk_index) + i)
                )

            expanded_results[f"Match on {id}"] = self.vector_collection.get(
                ids=expanded_ids,
            )["documents"]

        cleaned_results = {}
        # append docs together two at a time, removing overlap
        for match, docs in expanded_results.items():
            combined_result = ""
            # combined docs together
            if len(docs) == 0:
                combined_result = None
            elif len(docs) == 1:
                combined_result = docs[0]
            else:
                # combine first two docs, removing overlap, to
                # start the combined result
                substring_match = SequenceMatcher(
                    None, docs[0], docs[1]
                ).find_longest_match(0, len(docs[0]), 0, len(docs[1]))
                combined_docs = (
                    docs[0][: substring_match.a] + docs[1][substring_match.b :]
                )
                combined_result += combined_docs
                # eat these first two docs
                if len(docs) >= 3:
                    docs = docs[2:]
                else:
                    docs = []
                # continue eating the rest one by one
                while len(docs) >= 1:
                    substring_match = SequenceMatcher(None, combined_result, docs[0])

                    substring_match = substring_match.find_longest_match(
                        0, len(combined_result), 0, len(docs[0])
                    )

                    combined_result = (
                        combined_result[: substring_match.a]
                        + docs[0][substring_match.b :]
                    )

                    # eat the processed doc
                    if len(docs) >= 2:
                        docs = docs[1:]
                    else:
                        docs = []

                cleaned_results[match] = "..." + combined_result + "..."

        return cleaned_results


# --------------------END PAPER OBJECT SCHEMA-------------------------------



"""
ScrapeMed's PaperSet Module
============================

Module for building PMC paper datasets, known as paperSets.
This is the main endpoint for ScrapeMed users like data scientists and data engineers.

Functions for structured data generation, scraping PMC via both PMCID lists and
advanced PMC term searches.

"""

import scrapemed.scrape as scrape
from scrapemed.paper import Paper
import pandas as pd
from typing import Union, List
import matplotlib.pyplot as plt
from wordcloud import WordCloud


class paperSet:
    """
    A collection of Paper objects with various operations for managing and
    analyzing them.

    :param List[Paper] papers: A list of Paper objects to initialize the paperSet.

    This class represents a collection of Paper objects and provides methods
    for creating, adding, and visualizing papers.

    Methods:
    - from_search(email, term, retmax=10, verbose=False,
        suppress_warnings=True, suppress_errors=True): Generate a paperSet
        via a PMC search.
    - from_pmcid_list(pmcids, email, download=False, validate=True,
        strip_text_styling=True, verbose=False, suppress_warnings=True,
        suppress_errors=True): Generate a paperSet via a list of PMCIDs.
    - to_df(): Return a pandas DataFrame representation of the paperSet.
    - add_paper(paper): Add a Paper to the paperSet.
    - add_papers(papers): Add multiple Papers to the paperSet.
    - add_pmcid(pmcid, email, download=False, validate=True,
        strip_text_styling=True, verbose=False, suppress_warnings=True,
        suppress_errors=True): Add a Paper to the paperSet via a PMCID.
    - add_pmcids(pmcids, email, download=False, validate=True,
        strip_text_styling=True, verbose=False, suppress_warnings=True,
        suppress_errors=True): Add Papers to the paperSet via a list of PMCIDs.
    - visualize(): Generate a general visualization of the paperSet.
    - visualize_unique_values(columns_to_visualize=["Last_Updated",
        "Journal_Title"]): Visualize unique values in specified columns.
    - visualize_title_wordcloud(): Visualize a word cloud of all the
        Paper titles in the paperSet.
    """

    def __init__(self, papers: List[Paper]):
        """
        Initalize a paperSet. Usually called via `paperSet.from_search()` or
        `paperSet.from_pmcid_list()`.
        """
        self.papers = []
        for paper in papers:
            if paper:
                self.papers.append(paper)

        # Make a df of the papers
        paper_series_list = [paper.to_relational() for paper in self.papers]
        self.df = pd.DataFrame(paper_series_list)

        self.index = 0

        print("Done generating paperSet!")

        return None

    @classmethod
    def from_search(
        cls,
        email: str,
        term: str,
        retmax: int = 10,
        verbose: bool = False,
        suppress_warnings: bool = True,
        suppress_errors: bool = True,
    ):
        """
        Generate a paperSet via a PMC search.

        :param str email: Use your email to authenticate with PMC.
        :param str term: Search term.
        :param int retmax: Maximum number of PMCIDs to return (default is 10).
        :param bool verbose: Whether to display verbose output (default is False).
        :param bool suppress_warnings: Whether to suppress warnings while
            parsing XML (default is True).
        :param bool suppress_errors: Whether to return None on failed XML
            parsing, instead of raising an error (default is True).

        :returns: A paperSet generated from the PMC search results.
        :rtype: paperSet
        """
        print(
            (
                "Generating paperSet from search (This can take a while due to "
                "PMC HTTP Request Limitations!)..."
            )
        )
        pmcid_list = scrape.search_pmc(
            email=email, term=term, retmax=retmax, verbose=verbose
        )["IdList"]
        paper_list = [
            Paper.from_pmc(
                pmcid,
                email=email,
                verbose=verbose,
                suppress_warnings=suppress_warnings,
                suppress_errors=suppress_errors,
            )
            for pmcid in pmcid_list
        ]

        return cls(papers=paper_list)

    @classmethod
    def from_pmcid_list(
        cls,
        pmcids: List[int],
        email: str,
        download: bool = False,
        validate: bool = True,
        strip_text_styling: bool = True,
        verbose: bool = False,
        suppress_warnings: bool = True,
        suppress_errors: bool = True,
    ):
        """
        Generate a paperSet via a list of PMCIDs.

        :param List[int] pmcids: List of PMCIDs to populate the paperSet.
        :param str email: Use your email to authenticate with PMC.
        :param bool download: Whether or not to download the XMLs corresponding
            to PMCIDs (default is False).
        :param bool validate: Whether or not to validate the XMLs corresponding
            to PMCIDs (default is True).
        :param bool strip_text_styling: Whether or not to clean common HTML and
            other text styling out of the XMLs (default is True).
        :param bool verbose: Whether to display verbose output (default is False).
        :param bool suppress_warnings: Whether to suppress warnings while
            parsing XML (default is True).
        :param bool suppress_errors: Whether to return None on failed XML
            parsing, instead of raising an error (default is True).

        :returns: A paperSet generated from the list of PMCIDs.
        :rtype: paperSet
        """
        print(
            (
                "Generating paperSet from PMCID list (This can take a while due "
                "to PMC HTTP Request Limitations!)..."
            )
        )

        xml_list = scrape.get_xmls(
            pmcids=pmcids,
            email=email,
            download=download,
            validate=validate,
            strip_text_styling=strip_text_styling,
            verbose=verbose,
        )
        paper_list = [
            Paper.from_xml(
                pmcid,
                xml_root,
                verbose=verbose,
                suppress_warnings=suppress_warnings,
                suppress_errors=suppress_errors,
            )
            for pmcid, xml_root in zip(pmcids, xml_list)
        ]
        return cls(papers=paper_list)

    def __iter__(self):
        """
        Implement iteration for the paperSet.
        """
        return self

    def __next__(self):
        """
        Get the next paper when iterating over the paperSet.
        """
        if self.index < len(self.papers):
            result = self.papers[self.index]
            self.index += 1
            return result
        else:
            raise StopIteration

    def __len__(self):
        """
        Get the number of papers in the paperSet.
        """
        return len(self.papers)

    def __getitem__(self, index):
        """
        Get a paper from the paperSet by index.
        """
        if 0 <= index < len(self.papers):
            return self.papers[index]
        else:
            raise IndexError("Index out of range")

    def to_df(self):
        """
        Return a pandas DataFrame representation of the paperSet.

        :returns: DataFrame of the paperSet.
        :rtype: pd.DataFrame
        """
        return self.df

    def add_paper(self, paper: Paper):
        """
        Add a Paper to the paperSet directly. Returns True if the paper was
        added, False if the paper was already in the paperSet.

        :param Paper paper: The Paper to add to the paperSet.
        :returns: True if the paper was added, False if it was already in
            the paperSet.
        :rtype: bool
        """
        if paper not in self.papers:
            # caution: comparison of papers is sketchy! Be careful to not
            # duplicate papers in your paperSet
            self.papers.append(paper)
            new_row = paper.to_relational()
            self.df = pd.concat([self.df, new_row.to_frame().T], ignore_index=True)
            return True
        print(f"Paper with pmcid={paper.pmcid} already in paperSet.papers.")
        return False

    def add_papers(self, papers: List[Paper]):
        """
        Add Papers to the paperSet directly. Returns the number of papers added.

        :param List[Paper] papers: List of Papers to add to the paperSet.
        :returns: The number of papers added.
        :rtype: int
        """
        count_added = 0
        for paper in papers:
            if self.add_paper(paper):
                count_added += 1
        return count_added

    def add_pmcid(
        self,
        pmcid: Union[int, str],
        email: str,
        download: bool = False,
        validate: bool = True,
        strip_text_styling: bool = True,
        verbose: bool = False,
        suppress_warnings: bool = True,
        suppress_errors: bool = True,
    ):
        """
        Add a Paper to the paperSet via PMCID. Returns True if the paper was
        added, False if it was already in the paperSet.

        :param Union[int, str] pmcid: The PMCID to generate the Paper to add.
        :param str email: Use your email to authenticate with PMC.
        :param bool download: Whether or not to download the XMLs corresponding
            to pmcid (default is False).
        :param bool validate: Whether or not to validate the XMLs corresponding
            to pmcid (default is True).
        :param bool strip_text_styling: Whether or not to clean common HTML
            and other text styling out of the XML (default is True).
        :param bool verbose: Whether to display verbose output (default is False).
        :param bool suppress_warnings: Whether to suppress warnings while
            parsing XML (default is True).
        :param bool suppress_errors: Whether to return None on failed XML
            parsing, instead of raising an error (default is True).
        :returns: True if the paper was added, False if it was already in
            the paperSet.
        :rtype: bool
        """
        paper = Paper.from_pmc(
            pmcid,
            email,
            download=download,
            validate=validate,
            verbose=verbose,
            suppress_warnings=suppress_warnings,
            suppress_errors=suppress_errors,
        )
        return self.add_paper(paper)

    def add_pmcids(
        self,
        pmcids: List[Union[int, str]],
        email: str,
        download: bool = False,
        validate: bool = True,
        strip_text_styling: bool = True,
        verbose: bool = False,
        suppress_warnings: bool = True,
        suppress_errors: bool = True,
    ):
        """
        Add Papers to the paperSet via a list of PMCIDs. Returns the number
        of papers added.

        :param List[Union[int, str]] pmcids: List of PMCIDs to populate
            the paperSet.
        :param str email: Use your email to authenticate with PMC.
        :param bool download: Whether or not to download the XMLs
            corresponding to pmcids (default is False).
        :param bool validate: Whether or not to validate the XMLs
            corresponding to pmcids (default is True).
        :param bool strip_text_styling: Whether or not to clean common HTML
            and other text styling out of the XMLs (default is True).
        :param bool verbose: Whether to display verbose output (default is False).
        :param bool suppress_warnings: Whether to suppress warnings while
            parsing XML (default is True).
        :param bool suppress_errors: Whether to return None on failed XML
            parsing, instead of raising an error (default is True).
        :returns: The number of papers added.
        :rtype: int
        """
        count_added = 0
        for pmcid in pmcids:
            if self.add_pmcid(
                pmcid,
                email,
                download=download,
                validate=validate,
                verbose=verbose,
                suppress_warnings=suppress_warnings,
                suppress_errors=suppress_errors,
            ):
                count_added += 1
        return count_added

    def visualize(self):
        """
        Generates a general visualization of the paperSet, including unique
        value visualization and a wordcloud for all of the Paper titles.
        """
        self.visualize_unique_values()
        self.visualize_title_wordcloud()

        return None

    def visualize_unique_values(
        self, columns_to_visualize=["Last_Updated", "Journal_Title"]
    ) -> None:
        """
        Visualize unique values in specified columns of the paperSet DataFrame.

        :param List[str] columns_to_visualize: A list of column names to
            visualize (default is ["Last_Updated", "Journal_Title"]).
        """
        for column in columns_to_visualize:
            if column not in self.df.columns:
                print(f"Column '{column}' not found in paperSet.df.")
                continue

            unique_values = self.df[column].unique()
            value_counts = self.df[column].value_counts()

            if len(unique_values) <= 20:
                # Create a pie chart for columns with a small number
                # of unique values
                plt.figure(figsize=(6, 6))
                plt.pie(value_counts, labels=value_counts.index, autopct="%1.1f%%")
                plt.title(f"Pie Chart for {column}")
                plt.show()
            else:
                # Create a histogram for columns with many unique values
                plt.figure(figsize=(10, 6))
                plt.hist(self.df[column], bins=20, edgecolor="k")
                plt.title(f"Histogram for {column}")
                # can add labels, but is very messy
                # plt.xlabel(column)
                plt.ylabel("Frequency")
                plt.show()

        return None

    def visualize_title_wordcloud(self) -> None:
        """
        Visualize a wordcloud of all the Paper titles in the paperSet
        """

        titles = [p.title for p in self.papers]
        text = " ".join(titles)

        # Create a WordCloud object
        wordcloud = WordCloud(width=800, height=400, background_color="white").generate(
            text
        )

        # Display the word cloud using matplotlib
        plt.figure(figsize=(10, 5))
        plt.imshow(wordcloud, interpolation="bilinear")
        plt.axis("off")
        plt.title("Wordcloud of Paper Titles", fontsize=16)
        plt.show()

        return None

    # TODO: Add deletion methods if requested by ScrapeMed users.



"""
ScrapeMed's ``_text`` Module
==============================

The ``_text`` module of ScrapeMed is designed for organizing text found in
markup languages.

In the context of the ScrapeMed package, this module is used to facilitate the
organization of text found within paragraph (``<p>``) and section (``<sec>``)
tags in downloaded XML from PubMedCentral (PMC).

.. warnings::

   - :class:`multipleTitleWarning`: Warned when one title is expected but
        multiple are found.
   - :class:`unhandledTextTagWarning`: Warned when a tag is encountered in a
        text section of the XML but is not explicitly handled by ScrapeMed.
        The tag contents will be ignored if not handled manually. Feel free
        to submit a PR if you add a non-breaking code addition to handle these
        types of tags.
   - :class:`readHTMLFailure`: Warned when the pandas `read_html` function
        fails (rare). This may happen with tables void of readable data,
        such as tables that contain only graphics.
"""


import lxml.etree as ET
import textwrap
import scrapemed._clean as _clean
from scrapemed.utils import basicBiMap
import scrapemed._morehtml as mhtml
from itertools import chain
import warnings
import pandas as pd


# -------------------------------Warnings----------------------------
class multipleTitleWarning(Warning):
    """
    Warned when one title expected, but multiple found.
    """

    pass


class unhandledTextTagWarning(Warning):
    """
    Warned when a tag is encountered in a text section of the XML,
    but is not explicitly handled by ScrapeMed.

    The tag contents will be ignored if not handled manually.

    Feel free to submit a PR if you add a non-breaking code addition to handle
    these types of tags.
    """

    pass


class readHTMLFailure(Warning):
    """
    Warned when pandas read_html function fails (rare). May happen with
    tables void of readable data (ie. tables that contain only graphics).
    """

    pass


# ----------------------TextElement---------------------------------
class TextElement:
    """
    Base class for elements parsed from XML/HTML markup language.

    This class is initialized with a root element of the XML text element,
    as well as a reference map to populate with references found in the text,
    (and used to generate the replacement MHTML tags).

    :param ET.Element root: The root element of the XML text element.
    :param TextElement parent: The parent TextElement if applicable.
    :param basicBiMap ref_map: The reference map for storing references
        found in the text.

    Attributes:
        - root (ET.Element): The root element of the XML text element.
        - parent (TextElement, optional): The parent TextElement if applicable.
        - ref_map (basicBiMap): The reference map for storing references
            found in the text.

    Methods:
        - get_ref_map(): Return the shared BiMap for reference data.
        - set_ref_map(ref_map: basicBiMap): Set the shared BiMap for reference data.

    This class serves as the base class for more complex text classes.
    """

    def __init__(
        self,
        root: ET.Element,
        parent: "TextElement" = None,
        ref_map: basicBiMap = basicBiMap(),
    ):
        """
        Initialize a TextElement object.

        :param ET.Element root: The root element of the XML text element.
        :param TextElement parent: The parent TextElement if applicable.
        :param basicBiMap ref_map: The reference map for storing references
            found in the text
        """
        self.root = root
        self.parent = parent
        self.ref_map = ref_map

    # ------------------Getters and Setters for shared BiMap-------------------
    def get_ref_map(self) -> basicBiMap:
        """
        Return the shared BiMap for reference data.

        :returns: The shared BiMap containing reference data.
        :rtype: basicBiMap
        """
        ref_map = None
        if self.parent:
            ref_map = self.parent.get_ref_map()
        else:
            ref_map = self.ref_map
        return ref_map

    def set_ref_map(self, ref_map: basicBiMap):
        """
        Set the shared BiMap for reference data.

        :param basicBiMap ref_map: The BiMap containing reference data to be set.
        """
        if self.parent:
            self.parent.set_ref_map(ref_map)
        else:
            self.ref_map = ref_map
        return None

    # --------------End Getters and Setters for shared BiMap-----------------


# -------------------end TextElement-----------------------------------------


# ---------------------------TextParagraph----------------------------
class TextParagraph(TextElement):
    """
    Class representation of the data found in an XML <p> tag.

    :param ET.Element p_root: The root element of the XML <p> tag.
    :param TextElement parent: The parent TextElement if applicable.
    :param basicBiMap ref_map: The reference map for storing references
        found in the text.

    Attributes:
        - id (str): The identifier for the <p> tag.
        - text_with_refs (str): The text content of the <p> tag with references.
        - text (str): The clean text content of the <p> tag without references.

    Methods:
        - __str__(): Return the clean text content as a string.
        - __eq__(other): Check if two TextParagraph objects are equal based
            on their text content.
    """

    def __init__(
        self, p_root: ET.Element, parent=None, ref_map: basicBiMap = basicBiMap()
    ):
        """
        Initialize a TextParagraph object.

        :param ET.Element p_root: The root element of the XML <p> tag.
        :param TextElement parent: The parent TextElement if applicable.
        :param basicBiMap ref_map: The reference map for storing references
            found in the text.
        """
        super().__init__(
            root=p_root, parent=parent, ref_map=ref_map
        )  # initialize TextElement

        self.id = p_root.get("id")
        p_subtree = stringify_children(self.root)

        # split text and HTML tag references
        self.text_with_refs = _clean.split_text_and_refs(
            tree_text=p_subtree,
            ref_map=self.get_ref_map(),
            id=self.id,
            on_unknown="keep",
        )
        self.text = mhtml.remove_mhtml_tags(
            self.text_with_refs
        )  # clean text for str() and printing

    def __str__(self):
        """
        Return the clean text content of the TextParagraph object as a string.

        :returns: The clean text content.
        :rtype: str
        """
        return self.text

    def __eq__(self, other):
        """
        Check if two TextParagraph objects are equal based on their text
        content (including refs).

        :param TextParagraph other: The other TextParagraph object to compare with.

        :returns: True if the text content is equal, False otherwise.
        :rtype: bool
        """
        return self.text_with_refs == other.text_with_refs


# ------------------------end TextParagraph----------------------


# ----------------------------------TextSection------------------------------
class TextSection(TextElement):
    """
    Class representation of the data found in an XML <sec> tag.

    :param ET.Element sec_root: The root element of the XML <sec> tag.
    :param TextElement parent: The parent TextElement if applicable.
    :param basicBiMap ref_map: The reference map for storing references found
        in the text.

    Attributes:
        - title (str): The title of the section if available.
        - children (list): A list of TextSection, TextParagraph, TextTable, or
            TextFigure objects representing subsections, paragraphs, tables,
            or figures within the section.
        - text (str): The clean text content of the section without references.
        - text_with_refs (str): The text content of the section with references.

    Methods:
        - __str__(): Return a string representation of the section with proper
            indentation.
        - get_section_text(): Get a text representation of the entire section
            without references.
        - get_section_text_with_refs(): Get a text representation of the entire
            section with references.
        - __eq__(other): Check if two TextSection objects are equal based on
            their title and children.
    """

    def __init__(
        self, sec_root: ET.Element, parent=None, ref_map: basicBiMap = basicBiMap()
    ):
        """
        Initialize a text section from the root of a <sec> tree.

        Optionally provide reference BiMap storing links to reference data.
        Reference BiMap used to link text to table & figure data.

        Handles <title>, <p>, and <sec> children under the root <sec> tag.
        Anything else will be ignored and raise a warning.

        :param ET.Element sec_root: The root element of the XML <sec> tag.
        :param TextElement parent: The parent TextElement if applicable.
        :param basicBiMap ref_map: The reference map for storing references
            found in the text.
        """
        # initialize root, parent, and bimaps
        super().__init__(root=sec_root, parent=parent, ref_map=ref_map)

        self.title = None

        self.children = []  # section and paragraph children
        for child in sec_root.iterchildren():
            if child.tag == "title":
                if self.title:
                    warnings.warn(
                        multipleTitleWarning(
                            (
                                "Warning: Multiple Titles found for a single "
                                "TextSection. Check markup file formatting. "
                                "Using first title found."
                            )
                        )
                    )
                    continue
                self.title = child.text
            elif child.tag == "sec":
                self.children.append(
                    TextSection(child, parent=self, ref_map=self.get_ref_map())
                )
            elif child.tag == "p":
                self.children.append(
                    TextParagraph(child, parent=self, ref_map=self.get_ref_map())
                )
            elif child.tag == "table-wrap":
                self.children.append(
                    TextTable(child, parent=self, ref_map=self.get_ref_map())
                )
            elif child.tag == "fig":
                self.children.append(
                    TextFigure(child, parent=self, ref_map=self.get_ref_map())
                )
            else:
                warnings.warn(
                    (
                        f"Warning! Unexpected child with of type {child.tag} "
                        "found under an XML <sec> tag."
                    ),
                    unhandledTextTagWarning,
                )

        # after building the TextSection tree, get textual representations
        self.text = self.get_section_text()
        self.text_with_refs = self.get_section_text_with_refs()

    def __str__(self):
        """
        Return a string representation of the TextSection.

        The representation begins with the title if the section has one.
        Subsections will be indented, and body text will be printed without indentation.

        :returns: The string representation of the section.
        :rtype: str
        """
        s = ""
        if self.title is not None:
            s += f"SECTION: {self.title}:\n"
        for child in self.children:
            if isinstance(child, TextSection):
                s += "\n" + textwrap.indent(str(child), " " * 4)
                s += "\n"
            elif isinstance(child, TextParagraph):
                s += "\n" + str(child)
                s += "\n"
        return s

    def get_section_text(self):
        """
        Get a text representation of the entire text section, without references.

        :returns: The text content of the section.
        :rtype: str
        """
        return str(self)

    def get_section_text_with_refs(self):
        """
        Get a text representation of the entire text section, with references.

        :returns: The text content of the section with references.
        :rtype: str
        """
        s = ""
        if self.title is not None:
            s += f"SECTION: {self.title}:\n"
        for child in self.children:
            if isinstance(child, TextSection):
                s += "\n" + textwrap.indent(child.get_section_text_with_refs(), " " * 4)
                s += "\n"
            elif isinstance(child, TextParagraph):
                s += "\n" + child.text_with_refs
                s += "\n"
        return s

    def __eq__(self, other):
        """
        Check if two TextSection objects are equal based on their title and children.

        :param TextSection other: The other TextSection object to compare with.
        :returns: True if the titles and children are equal, False otherwise.
        :rtype: bool
        """

        # TODO: Test if this actually works for comparing TextSections.
        return self.title == other.title and self.children == other.children


# ----------------------end TextSection----------------------------------


# ---------------------------TextTable-------------------------------------
class TextTable(TextElement):
    """
    Initialize and process a table-wrap found in a text element of PMC XML.

    Uses pandas' `read_html` function (which relies on lxml and falls
    back to html5lib) to process the HTML tables into dataframes.

    Adds labels and captions if notated in the XML under
    //table-wrap/label and //table-wrap/caption/p tags.

    :param ET.Element table_root: The root element of the table-wrap found in PMC XML.
    :param TextElement parent: The parent TextElement if applicable.
    :param basicBiMap ref_map: The reference map for storing references found
        in the text.

    Attributes:
        - df (pandas.DataFrame): The dataframe representation of the table.
    """

    def __init__(
        self, table_root: ET.Element, parent=None, ref_map: basicBiMap = basicBiMap()
    ):
        """
        Initialize and process table-wrap found in a text element of PMC XML.

        Use pandas' read_html function (which relies on lxml and falls
        back to html5lib) to process the HTML tables into dataframes.

        Adds labels and captions if notated in the XML under
        //table-wrap/label and //table-wrap/caption/p tags.

        :param ET.Element table_root: The root element of the table-wrap found
            in PMC XML.
        :param TextElement parent: The parent TextElement if applicable.
        :param basicBiMap ref_map: The reference map for storing references
            found in the text.
        """
        # initialize root, parent, and bimaps
        super().__init__(root=table_root, parent=parent, ref_map=ref_map)

        # find label if any
        label_matches = table_root.xpath("label")
        label = None
        if len(label_matches) > 0:
            label = label_matches[0].text
        # find caption if any
        caption_matches = table_root.xpath("caption/p")
        caption = None
        if len(caption_matches) > 0:
            caption = caption_matches[0].text

        table_xml_str = ET.tostring(table_root)
        try:
            table_df = pd.read_html(table_xml_str)[0]
        except ValueError:
            warnings.warn(
                (
                    f"Table with label {label} and caption {caption} could not "
                    "be parsed with pd.read_html."
                ),
                readHTMLFailure,
            )
            self.df = None
            return None

        # build title for styled df, if relevant
        title = None
        if label and caption:
            title = f"{label}: {caption}"
        elif label:
            title = label
        elif caption:
            title = caption

        if title:
            table_df = table_df.style.set_caption(title)

        self.df = table_df

        return None

    def __str__(self):
        """
        Return a string representation of the table using member `.df`.

        :returns: The string representation of the table.
        :rtype: str
        """
        return str(self.df)

    def __repr__(self):
        """
        Return a string representation of the table using member `.df`.

        :returns: The string representation of the table.
        :rtype: str
        """
        return repr(self.df)


# ---------------------end TextTable-------------------------------------


# ---------------------------TextFigure-----------------------------------
class TextFigure(TextElement):
    """
    Initialize and parse a figure found in a text element of a PMC XML.

    Parses figures into a dictionary with their information
    (label, caption, and link).
    Unfortunately, the links are relative and cannot be reliably
    traced to a public URI.
    This means I have not found a way to download the actual
    figures to store via Pillow, etc.

    :param ET.Element fig_root: The root element of the figure found in PMC XML.
    :param TextElement parent: The parent TextElement if applicable.
    :param basicBiMap ref_map: The reference map for storing references found
        in the text.

    Attributes:
        - fig_dict (dict): A dictionary containing figure information with keys:
            - 'Label': The label of the figure.
            - 'Caption': The caption of the figure.
            - 'Link': The link (relative) to the figure.
    """

    def __init__(
        self, fig_root: ET.Element, parent=None, ref_map: basicBiMap = basicBiMap()
    ):
        """
        Initialize and parse a figure found in a text element of a PMC XML.

        Parses figures into a dictionary with their information
        (label, caption, and link).
        Unfortunately, the links are relative and cannot be reliably
        traced to a public URI.
        This means I have not found a way to download the actual
        figures to store via Pillow, etc.

        :param ET.Element fig_root: The root element of the figure found in
            PMC XML.
        :param TextElement parent: The parent TextElement if applicable.
        :param basicBiMap ref_map: The reference map for storing references
            found in the text.
        """

        # TODO: Find a way to grab actual figures. May be impossible with PMC
        # based on the research I've done so far.

        # initialize root, parent, and bimaps
        super().__init__(root=fig_root, parent=parent, ref_map=ref_map)

        root = fig_root
        label = root.find(".//label")
        if label is not None:
            label = label.text
        caption = root.find(".//caption")
        if caption is not None:
            caption = "".join(caption.itertext())
        graphic_href = root.find(".//graphic").get("{http://www.w3.org/1999/xlink}href")

        fig_dict = {"Label": label, "Caption": caption, "Link": graphic_href}

        self.fig_dict = fig_dict

        return None

    def __str__(self):
        """
        Return a string representation of the figure based on dict info in `.fig_dict`.

        :returns: The string representation of the figure.
        :rtype: str
        """
        return str(self.fig_dict)

    def __repr__(self):
        """
        Return a string representation of the figure  based on dict info in `.fig_dict`.

        :returns: The string representation of the figure.
        :rtype: str
        """
        return repr(self.fig_dict)


# --------------------------end TextFigure----------------------------------


# -------------------------------End Classes----------------------------


# ---------------------------------Helpers---------------------------------
def stringify_children(node, encoding="utf-8"):
    """
    Returns a string representation of a node and all its children
    (recursively), including markup language tags.

    Turns any byte strings in the subtree representation to regular strings,
    following the provided encoding.

    :param node: The XML node to stringify.
    :type node: ET.Element
    :param str encoding: The encoding to use for decoding byte strings
        (default is 'utf-8').

    :returns: A string representation of the node and its children,
        including markup language tags.
    :rtype: str
    """
    subtree = [
        chunk
        for chunk in chain(
            (node.text,),
            chain(
                *(
                    (ET.tostring(child, with_tail=False), child.tail)
                    for child in node.getchildren()
                )
            ),
            (node.tail,),
        )
        if chunk
    ]
    # decode any bytestrings
    for i in range(len(subtree)):
        if isinstance(subtree[i], bytes):
            subtree[i] = subtree[i].decode(encoding)
    return "".join(subtree).strip()



"""
ScrapeMed's Markup Language Cleaning Utilities
===============================================

Scrapemed module for markup language cleaning utilities.

.. warns::

   unexpectedTagWarning
       Warned when an unexpected tag enclosed in angle brackets is found.

"""

import warnings
import re
from scrapemed.utils import basicBiMap
import scrapemed._morehtml as mhtml

# monkeypatch warnings.formatwarning for cleaner warnings
warnings.formatwarning = (
    lambda msg, category, *args, **kwargs: f"{category.__name__}: {msg}\n\n"
)


class unexpectedTagWarning(Warning):
    """
    Warned when an unexpected tag enclosed in angle brackets is found.
    """

    pass


def clean_xml_string(xml_string: str, strip_text_styling=True, verbose=False):
    """
    Clean an XML string.

    :param str xml_string: The XML string to be cleaned.
    :param bool strip_text_styling: Whether to remove or replace HTML text styling tags.
    :param bool verbose: Whether to print verbose output.

    :return: The cleaned XML string.
    :rtype: str
    """
    # Strip html styling if requested
    if strip_text_styling:
        xml_string = _remove_text_styling(xml_string, verbose=verbose)
    return xml_string


def _remove_text_styling(text: str, verbose=False) -> str:
    """
    Remove specified HTML stylings from the provided text.

    This function removes specified HTML stylings from the input text. It can
    remove opening tags and their corresponding closing tags and replace
    specific opening tags with desired values.

    **Specifically:**
    This overloaded wrapper function removes italic, bold, underline HTML text
    styling tags from the input text. Additionally, it replaces <sub> with "_"
    and <sup> with "^". <ext-link> is replaced with "[External URI:]".

    :param str text: The text containing HTML stylings to be removed or replaced.
    :param list[str] removals: A list of opening tags to be removed. Their
        corresponding closing tags will also be removed. Tags will be removed
        regardless of attributes.
    :param dict replaces: A dictionary of find, replace values. The find values
        should be HTML opening tags. They will be matched regardless of attributes.
    :param bool verbose: Whether to print verbose output.

    :return: The XML string with default HTML text styling tags (`REMOVALS`, `REPLACES`)
        removed or replaced.
    :rtype: str
    """

    # remove italic, bold, underline styling
    REMOVALS = ["<italic>", "<i>", "<bold>", "<b>", "<underline>", "<u>"]
    REPLACES = {"<sub>": "_", "<sup>": "^", "<ext-link>": "[External URI:]"}

    return _remove_html_styling(
        text, removals=REMOVALS, replaces=REPLACES, verbose=verbose
    )


def _remove_html_styling(
    text: str, removals: list[str], replaces: dict, verbose=False
) -> str:
    """
    Remove specified HTML stylings from the provided text.

    :param str text: The text containing HTML stylings to be removed.
    :param list[str] removals: A list of opening tags to be removed. Their
        corresponding closing tags will also be removed. Tags will be removed
        regardless of attributes.
    :param dict replaces: A dictionary of find, replace values. The find values
        should be HTML opening tags. They will be matched regardless of attributes.
    :param bool verbose: Whether to print verbose output.

    :return: The XML string with specified HTML text styling tags removed.
    :rtype: str
    """

    # ADD IN CLOSING TAGS FOR REMOVAL TAGS
    to_remove = removals.copy()
    more_to_remove = []
    for tag in to_remove:
        more_to_remove.append(tag[0] + "/" + tag[1:])
    to_remove.extend(more_to_remove)
    # ADD IN CLOSING TAGS FOR REPLACEMENT TAGS
    to_replace_basic = replaces.copy()
    for tag in to_replace_basic.keys():
        to_remove.append(tag[0] + "/" + tag[1:])

    # MATCH REGARDLESS OF HTML ATTRIBUTES
    # Sample of what removals should look like for tag matching
    # regardless of attributes
    # /<head\b[^>]*>/i
    for i in range(len(to_remove)):
        to_remove[i] = to_remove[i][0:-1] + "\\b[^>]*" + to_remove[i][-1]
    to_replace = {}
    for find, replace in to_replace_basic.items():
        new_find = find[0:-1] + "\\b[^>]*" + find[-1]
        to_replace[new_find] = replace

    # REPORT REQUESTED BEHAVIOR AT RUNTIME
    if verbose:
        print(f"Removing the following tags:\n{to_remove}\n")
        print("Making the following replacements:\n")
        for find, replace in to_replace.items():
            print(f"{find} replaced with {replace}\n")

    # REMOVALS
    removal_pattern = "|".join(to_remove)
    r = re.compile(removal_pattern, re.IGNORECASE)
    text = r.sub("", text)

    # REPLACEMENTS
    for find, replace in to_replace.items():
        text = re.sub(find, replace, text)

    # RETURN THE CLEANED TEXT
    return text


def split_text_and_refs(
    tree_text: str, ref_map: basicBiMap, id=None, on_unknown="keep"
):
    """
    Split HTML tags out of text.

    - HTML text styling tags will be removed if they aren't already.
    - <xref>, <table-wrap>, and <fig> tags will be converted to MHTML tags containing
      the key to use when searching for these references, tables, and figures.

    Returns the cleaned text and updates the passed BiMap for any new key-tag
    pairs found.

    :param str tree_text: A string representing a markup language tree containing
        HTML tags.
    :param ref_map: A BiMap containing keys connected to reference tag values. BiMap
        forward keys should be reference keys to place into the text in lieu of the
        tag for later BiMap table lookup. BiMap forward values should be the
        actual tags. The provided BiMap will be modified to reflect any new tag
        values found, and keys will be appended as necessary.
    :type ref_map: basicBiMap
    :param id: Optionally provide an id for traceback of any issues.
    :type id: Any, optional
    :param str on_unknown: Behavior when encountering an unknown tag. Determines
        what happens to the tag contents.
        Default is 'keep'. Options: ['drop', 'keep']

    :return: A tuple containing the cleaned text and the updated BiMap.
    :rtype: Tuple[str, basicBiMap]
    """
    XREF_TAG_NAME = "xref"
    FIGURE_TAG_NAME = "fig"
    TABLEWRAP_TAG_NAME = "table-wrap"
    ALLOWED_TAG_NAMES = [XREF_TAG_NAME, FIGURE_TAG_NAME, TABLEWRAP_TAG_NAME]

    # regex pattern string to match tags through to closing tag or self closing
    # should match any HTML or XML tag
    XML_HTML_TAG_PATTERN = (
        r"<([a-zA-Z][\w-]*)\b[^>]*>(.*?)</\1>|<([a-zA-Z][\w-]*)\b[^/>]*/?>"
    )
    tag_r = re.compile(
        XML_HTML_TAG_PATTERN, re.DOTALL
    )  # DOTALL used in case of multiline tag spans

    text = tree_text.strip()
    text = _remove_text_styling(text)
    cleaned_text = ""

    while len(text) > 0:
        match = tag_r.search(text)
        if match:
            # found a tag, append the text prior to the tag
            # and deal w tag
            # EAT NEXT TAG AND MATCH PARTS
            tag_name = match.group(1)
            tag_contents = match.group(2)
            full_tag = match.group()

            # ADD CONTENTS PRIOR TO TAG
            tag_start_index = match.start()
            cleaned_text += text[0:tag_start_index]

            # UNKNOWN TAG PROCESSING, WARN AND PERFORM SPECIFIED BEHAVIOR
            if tag_name not in ALLOWED_TAG_NAMES:
                warning_msg = (
                    f"Tag of type {tag_name} found in a text portion of "
                    "the provided markup language. "
                    "Expected only HTML styling tags, or tags from the "
                    f"following list: {ALLOWED_TAG_NAMES}."
                    f" Specified unknown tag behavior: {on_unknown}."
                )
                if id:
                    warning_msg += (
                        " Warning occured in a text section " f"with id: {id}."
                    )
                warnings.warn(warning_msg, unexpectedTagWarning)
                if on_unknown == "keep":
                    cleaned_text += tag_contents
                # eat through the text that was just processed
                text = text[match.end() :]

            # KNOWN TAG PROCESSING, UPDATE DATA REF
            else:
                # add tag contents if it is an xref.
                if tag_name == "xref":
                    cleaned_text += tag_contents
                # Get reference number for data reference
                ref_num = None
                if full_tag in ref_map.reverse:
                    # have we generated a map for this tag before?
                    ref_num = ref_map.reverse[full_tag]
                else:
                    ref_num = len(ref_map)  # new tag, append a new key
                    ref_map[ref_num] = full_tag  # and fill in the tag value

                data_ref_tag = mhtml.generate_typed_mhtml_tag(
                    tag_type="dataref", string=str(ref_num)
                )
                cleaned_text += f"{data_ref_tag}"

                # eat through the text that was just processed
                text = text[match.end() :]

        else:
            # no more tags to deal with, add the last bits to our output text
            cleaned_text += text
            text = ""

    return cleaned_text



"""
ScrapeMed's Scrape Module
============================

ScrapeMed's `scrape` module handles PubMed Central data searching
and downloads.

This module also handles conversion of raw XML data to
lxml.etree.ElementTree objects.

..warnings::
    - :class:`validationWarning` - Warned when downloading PMC XML without
        validating.
"""


import scrapemed._clean as _clean
import scrapemed._validate as _validate
import lxml.etree as ET
from Bio import Entrez
import warnings
from typing import List


class validationWarning(Warning):
    """
    Warned when downloading PMC XML without validating.
    """

    pass


# ---------------------Download Funcs for PubMed Central-----------------------
def search_pmc(email: str, term: str, retmax: int = 10, verbose: bool = False) -> dict:
    """
    Wrapper for Bio.Entrez's esearch function to retrieve PMC search results.

    :param str email: Use your email to authenticate with PMC.
    :param str term: The search term.
    :param int retmax: The maximum number of PMCIDs to return. Default is 10.
    :param bool verbose: Whether to display verbose output. Default is False.

    :return: A dictionary containing search results, including PMCIDs.
    :rtype: dict
    """

    DB = "pmc"
    Entrez.email = email
    handle = Entrez.esearch(db=DB, retmax=retmax, term=term, idtype="pmc")
    record = Entrez.read(handle)
    handle.close()

    if verbose:
        print(f"\nSearching {DB}...\n")
        print(f"Number of results found: {record['Count']}")

    return record


def get_xmls(
    pmcids: List[int],
    email: str,
    download=False,
    validate=True,
    strip_text_styling=True,
    verbose=False,
) -> List[ET.ElementTree]:
    """
    Retrieve XMLs of research papers from PMC, given a list of PMCIDs.
    Also validates and cleans the XMLs by default.

    :param List[int] pmcids: List of PMCIDs of articles to retrieve.
    :param str email: Use your email to authenticate with PMC.
    :param bool download: Whether or not to download the XMLs. Default is False.
    :param bool validate: Whether or not to validate the retrieved XMLs
        (HIGHLY RECOMMENDED). Default is True.
    :param bool strip_text_styling: Whether or not to clean common HTML text
        styling from the text (HIGHLY RECOMMENDED). Default is True.
    :param bool verbose: Whether to display verbose output. Default is False.

    :return: List of ElementTrees of the XMLs corresponding to
        the provided PMCIDs.
    :rtype: List[ET.ElementTree]
    """
    return [
        get_xml(pmcid, email, download, validate, strip_text_styling, verbose)
        for pmcid in pmcids
    ]


def get_xml(
    pmcid: int,
    email: str,
    download=False,
    validate=True,
    strip_text_styling=True,
    verbose=False,
) -> ET.ElementTree:
    """
    Retrieve XML of a research paper from PMC, given a PMCID.
    Also validates and cleans the XML by default.

    :param int pmcid: PMCID of the article to retrieve.
    :param str email: Use your email to authenticate with PMC.
    :param bool download: Whether or not to download the XML. Default is False.
    :param bool validate: Whether or not to validate the retrieved XML
        (HIGHLY RECOMMENDED). Default is True.
    :param bool strip_text_styling: Whether or not to clean common HTML
        text styling from the text (HIGHLY RECOMMENDED). Default is True.
    :param bool verbose: Whether to display verbose output. Default is False.

    :return: ElementTree of the validated XML record.
    :rtype: ET.ElementTree
    """
    xml_text = _get_xml_string(pmcid, email, download, verbose)
    tree = xml_tree_from_string(
        xml_string=xml_text, strip_text_styling=strip_text_styling, verbose=verbose
    )

    if validate:
        # Validate tags, attrs, values are supported for
        # parsing by the scrapemed package.
        _validate.validate_xml(tree)
    else:
        warnings.warn(
            (
                f"Warning! Scraping XML for PMCID {pmcid} from "
                "PMC without validating."
            ),
            validationWarning,
        )

    return tree


def _get_xml_string(pmcid: int, email: str, download=False, verbose=False) -> str:
    """
    Retrieve XML text of a research paper from PMC.

    :param int pmcid: PMCID of the article to retrieve.
    :param str email: Email of the user requesting data from PMC.
    :param bool download: Whether or not to download the XML. Default is False.
    :param bool verbose: Whether to display verbose output. Default is False.

    :return: XML Text of the record.
    :rtype: str

    WARNING: THIS FUNCTION DOES NOT VALIDATE THE XML.
    """
    DB = "pmc"
    RETTYPE = "full"
    RETMODE = "xml"
    Entrez.email = email

    # Actually fetch from PMC
    handle = Entrez.efetch(db=DB, id=pmcid, rettype=RETTYPE, retmode=RETMODE)
    xml_record = handle.read()
    xml_text = xml_record.decode(encoding="utf-8")
    handle.close()

    if verbose:
        print(f"\nGetting {RETMODE.upper()} string from {DB}...\n")
        print(f"XML Record First 100 bytes: {xml_record[0:100]}")
        print(f"XML Text First 100 Chars: {xml_text[0:100]}")

    if download:
        with open(f"data/entrez_download_PMCID={pmcid}.{RETMODE}", "w+") as f:
            f.write(xml_text)

    return xml_text


# ----------------End Download Funcs for PubMed Central---------------------


# --------------------Convert XML strings -> Trees---------------------
def xml_tree_from_string(
    xml_string: str, strip_text_styling, verbose=False
) -> ET.ElementTree:
    """
    Converts a string representing XML to an lxml ElementTree.

    :param str xml_string: A string or bytestream representing XML.
    :param bool strip_text_styling: Whether to remove HTML text styling tags or not.
    :param bool verbose: Whether to display verbose output. Default is False.

    :return: An lxml.etree.ElementTree of the passed string.
    :rtype: ET.ElementTree
    """
    xml_string = _clean.clean_xml_string(xml_string, strip_text_styling)
    tree = ET.ElementTree(ET.fromstring(xml_string))
    return tree


# --------------------End Convert XML strings -> Trees---------------------



"""
ScrapeMed's Utility Module
============================

The `utils` module contains various utilities necessary for PMC data parsing and
cleaning, where the utilities are not directly related to text cleaning (`_clean`),
recursive text processing (`_text`), or parsing (`_parse`).

At the moment, the module contains a helper function for cleaning up docstrings,
and a class, basicBiMap, which is a two-way map similar to python's dict class,
used for efficient storage of data reference maps used throughout ScrapeMed.

Note: Data reference maps are used to pull citations, tables, and figures out of
text for parsing elsewhere while retaining placeholders in the original text.

.warnings::
    - :class:`reversedBiMapComparisonWarning` - Warned when comparing
        basicBiMaps which are exactly the same but reversed.
"""

import warnings
from inspect import cleandoc


class reversedBiMapComparisonWarning(Warning):
    """
    Warned when comparing basicBiMaps which are exactly the same but reversed.
    """

    pass


# --------- general helper funcs ---------------
def cleanerdoc(s):
    """
    Wrapper for inspect.cleandoc which also removes newlines.

    :param str s: The string to clean and remove newlines from.

    :return: The cleaned and newline-removed string.
    :rtype: str
    """
    return cleandoc(s).replace("\n", "")


# --------- end general helper funcs


class basicBiMap(dict):
    """
    BiMap class which extends Python's dict class to also store
    a reverse of itself for efficiency.

    This class allows for bidirectional mapping between keys and values.
    When a key-value pair is added, it can be accessed both by key and by value.

    Methods are inherited from the dict class, and additional methods
    specific to bidirectional mapping are provided.

    :param *args: Positional arguments passed to the dict constructor.
    :param **kwargs: Keyword arguments passed to the dict constructor.

    Example:
    ```
    bi_map = basicBiMap({'apple': 'red', 'banana': 'yellow'})
    print(bi_map['apple'])  # Output: 'red'
    print(bi_map.reverse['red'])  # Output: 'apple'
    ```

    :ivar reverse: A reverse mapping from values to keys.
    :vartype reverse: dict
    """

    def __init__(self, *args, **kwargs):
        """
        Initialize the BiMap.

        :param *args: Positional arguments passed to the dict constructor.
        :param **kwargs: Keyword arguments passed to the dict constructor.
        """
        super().__init__(*args, **kwargs)
        self.reverse = {value: key for key, value in self.items()}

    def __str__(self):
        """
        Return a string representation of the BiMap.

        :return: A string representation of the BiMap.
        :rtype: str
        """
        return "BiMap: " + super().__str__()

    def __setitem__(self, key, value):
        """
        Set a key-value pair in the BiMap.

        :param key: The key to set.
        :param value: The value to set.
        """
        super().__setitem__(key, value)  # call the original __setitem__ method
        self.reverse[value] = key  # update the reverse dict

    # Pass to super for methods necessary to be considered a Mapping class
    def __getitem__(self, key):
        return super().__getitem__(key)

    def __iter__(self):
        return super().__iter__()

    def __next__(self):
        return super().__next__()

    def __len__(self):
        return super().__len__()

    def __eq__(self, other):
        """
        Compare the BiMap to another object for equality.

        :param other: The object to compare to.
        :type other: dict or basicBiMap

        :return: True if the BiMap is equal to the other object, False otherwise.
        :rtype: bool
        """
        if not isinstance(other, basicBiMap) and not issubclass(type(other), dict):
            return False
        if not super().__eq__(other):
            if isinstance(other, basicBiMap) and other.reverse == dict(self):
                warnings.warn(
                    (
                        "Warning: Comparing basicBiMaps which are not equal, "
                        "but would be with keys and values swapped."
                    ),
                    reversedBiMapComparisonWarning,
                )
            return False
        return True



"""
ScrapeMed's Trees Module
============================

Scrapemed's `trees` module handles PMC article tree visualizations,
statistics, and descriptions.
"""

import copy
from graphviz import Digraph
import lxml.etree as ET


def investigate_xml_tree(root: ET.Element) -> None:
    """
    Print basic statistics and information about an XML tree provided its root.

    :param ET.Element root: The root of an ElementTree of your XML.

    This function prints the following information to stdout:
    - Number of elements in the XML tree.
    - Unique element types in the XML tree.
    - A dictionary with tag frequencies in the XML tree.
    """

    elem_list = [elem.tag for elem in root.iter()]
    num_elements = len(elem_list)
    all_element_types = set(elem_list)

    # PRINT UNIQUE ELEMENTS
    print(f"Num Elements: {num_elements}")
    print(f"Unique Element Types: {all_element_types}")
    print("--------------------------------------------\n")

    # PRINT OVERALL DATA DICTIONARY
    print(f"Tag Dictionary:\n {_generate_tag_dictionary(root)}\n")
    print("--------------------------------------------\n")
    return


def visualize_element_tree(
    root: ET.Element, title="data/element_tree.gv", test_mode=False
) -> None:
    """
    Visualize an XML element tree using Graphviz.

    :param ET.Element root: The root of the XML element tree to visualize.
    :param str title: The title or filename for the output visualization.
        Default is "data/element_tree.gv".
    :param bool test_mode: Whether to render the visualization in test mode or not.

    This function creates a visualization of the XML element tree using Graphviz
    and optionally renders it.
    """
    root = copy.copy(root)
    dot = Digraph()
    _add_elements(dot, root)
    if not test_mode:
        dot.render(title, view=True)
    return


def _add_elements(dot: Digraph, element: ET.Element, parent=None):
    """
    Recursively add elements to a Graphviz dot graph.

    :param Digraph dot: The Graphviz dot graph.
    :param ET.Element element: The XML element to add to the graph.
    :param str parent: The parent element's tag. Default is None.

    This function is used internally to recursively add XML elements and their
    relationships to the Graphviz dot graph.
    """
    if parent is not None:
        dot.edge(parent, element.tag)
    dot.node(element.tag, element.tag)
    for child in element:
        _add_elements(dot, child, element.tag)
    return


def _generate_tag_dictionary(root: ET.Element) -> dict:
    """
    Generate a dictionary of all tags, each with a subdictionary of attributes,
    and lists of values seen for each attribute.

    This function helps define the scope of tags for a given XML, HTML, or
    other tagged language tree.

    :param ET.Element root: The root of an ElementTree of XML, HTML, or other
    tagged language.

    :return: A dictionary of tags, each with a subdictionary of attributes and
    lists of values for each attribute.
    :rtype: dict

    The structure of the overall dictionary is as follows:
    {
        tag_0:
        {
            attr_0: [value_0, ..., value_n],
            attr_1: [value_0, ..., value_n],
            ...
        },
        tag_1:
        {
            attr_0: [value_0, ..., value_n],
            attr_1: [value_0, ..., value_n],
            ...
        },
        ...
    }
    """
    data_dict = {}

    for element in root.iter():
        # ignore processing instructions
        if isinstance(element, ET._ProcessingInstruction):
            continue

        # dictionary of key-value pairs such that we have
        # {attr: list of values the attribute can take on}
        attr_values_dict = {}

        try:
            # grab existing attr-val dict if we've seen this tag before
            # grab reference to the existing attr-val dict,
            # intentionally not copied
            attr_values_dict = data_dict[element.tag]

            for attr, val in element.attrib.items():
                try:
                    # if weve seen the attribute before,
                    # add the value if its new
                    if val not in attr_values_dict[attr]:
                        attr_values_dict[attr].append(val)
                # we havent seen the attribute before at all
                except KeyError:
                    attr_values_dict[attr] = [val]
        except KeyError:
            # create new attr-val dict, we haven't seen this tag before
            data_dict[element.tag] = None
            # add attributes if they exist
            data_dict[element.tag] = copy.copy(element.attrib)
            # now put values into lists so they can be
            # appended to later if needed
            attr_values_dict = data_dict[element.tag]
            for attr, val in attr_values_dict.items():
                attr_values_dict[attr] = [attr_values_dict[attr]]

    return data_dict


# ----------------END DESCRIBE / CONVERT DATA-------------------



"""
ScrapeMed's ``_parse`` Module
=============================================

Parse module for grabbing metadata, text, tables, figures, etc.
from XML trees representing PMC articles.

DTD for the XML should be NLM articleset 2.0.
Otherwise the behavior here may not be as expected.

Middleman between the `scrape` module and the `paper` module for ScrapeMed.

..warnings::
    - :class:`unexpectedMultipleMatchWarning` - Warned when one match is
        expected, but multiple are found.
    - :class:`unexpectedZeroMatchWarning` - Warned when one or more matches are
        expected, and none are found.
    - :class:`badTextFormattingWarning` - Warned when there are issues with text
        formatting.
    - :class:`unmatchedCitationWarning` - Warned when a citation reference is
        made but not matched to an actual <ref> tag.
    - :class:`unmatchedTableWarning` - Warned when a table reference is made
        but not matched to an actual <table-wrap> tag.
    - :class:`unmatchedFigureWarning` - Warned when a figure reference is made
        but not matched to an actual <fig> tag.
"""

import copy
from typing import List, Dict, Tuple, Set
from typing import Union
import scrapemed.scrape as scrape
import lxml.etree as ET
from scrapemed.utils import basicBiMap, cleanerdoc
from scrapemed._text import TextParagraph, TextSection, TextTable, TextFigure
from datetime import datetime
import pandas as pd
import warnings
import textwrap
import uuid


# -----------Custom Warnings & Exceptions for Parsing------------
class unexpectedMultipleMatchWarning(Warning):
    """
    Warned when one match expected, but multiple found.
    """

    pass


class unexpectedZeroMatchWarning(Warning):
    """
    Warned when one or more matches expected, and none are found.
    """

    pass


class badTextFormattingWarning(Warning):
    pass


class unmatchedCitationWarning(Warning):
    """
    Warned when a citation reference is made but not matched to an
    actual <ref> tag.
    """

    pass


class unmatchedTableWarning(Warning):
    """
    Warned when a table reference is made but not matched to an
    actual <table-wrap> tag.
    """

    pass


class unmatchedFigureWarning(Warning):
    """
    Warned when a figure reference is made but not matched
    to an actual <fig> tag.
    """

    pass


# -----------End Custom Warnings & Exceptions for Parsing------------


# ------------------GENERATE PAPER DICTIONARY GIVEN PMCID--------------------
def paper_dict_from_pmc(
    pmcid: int,
    email: str,
    download: bool = False,
    validate: bool = True,
    verbose: bool = False,
    suppress_warnings: bool = False,
    suppress_errors: bool = False,
) -> dict:
    """
    Wrapper that scrapes a PMC article specified by PMCID from the web,
    then parses the retrieved XML into a dictionary of useful values.

    This function serves as a middleman between the `scrape.py` module and
    `Paper.from_pmc` method in `paper.py`, facilitating the conversion of
    PMC XML data to a dictionary.

    :param int pmcid: Unique PMCID for the article to scrape and parse.
    :param str email: Provide your email address for authentication with PMC.
    :param bool download: Whether or not to download the XML retrieved from PMC.
    :param bool validate: Whether or not to validate the XML from PMC against
        NLM articleset 2.0 DTD (HIGHLY RECOMMENDED).
    :param bool verbose: Whether or not to have verbose output for debugging.
    :param bool suppress_warnings: Whether to suppress warnings while parsing XML.
        Note: Warnings are frequent, given the variable nature of PMC XML data.
        Recommended to suppress when parsing many XMLs at once.
    :param bool suppress_errors: Return None on failed XML parsing, instead of
        raising an error.

    :return: A dictionary containing useful values parsed from the PMC article.
    :rtype: dict
    """

    if verbose:
        print(f"Generating Paper object for PMCID = {pmcid}...")
    # DOWNLOAD XML TREE AND GET ROOT
    paper_tree = scrape.get_xml(
        pmcid=pmcid, email=email, download=download, validate=validate, verbose=verbose
    )
    root = paper_tree.getroot()

    return generate_paper_dict(
        pmcid=pmcid,
        paper_root=root,
        verbose=verbose,
        suppress_warnings=suppress_warnings,
        suppress_errors=suppress_errors,
    )


def generate_paper_dict(
    pmcid: int,
    paper_root: ET.Element,
    verbose: bool = False,
    suppress_warnings: bool = False,
    suppress_errors: bool = False,
) -> dict:
    """
    Given the root of an XML tree, parse through it and generate
    a flattened dictionary of relevant PMC paper XML information.

    This function expects the XML to be in NLM articleset 2.0 DTD format.

    Optionally, you can suppress warnings and/or errors. If errors are suppressed,
    None will be returned upon failed parsing.

    :param int pmcid: Unique PMCID for the article being parsed.
    :param ET.Element paper_root: The root element of the PMC paper XML tree.
    :param bool verbose: Whether or not to have verbose output for debugging.
    :param bool suppress_warnings: Whether to suppress warnings while parsing XML.
        Note: Warnings are frequent due to the variable nature of PMC XML data.
        Recommended to suppress when parsing many XMLs at once.
    :param bool suppress_errors: Whether to suppress errors during parsing.
        If suppressed, None will be returned upon a failed parsing attempt.

    :return: A flattened dictionary containing relevant PMC paper XML information.
    :rtype: dict or None if errors are suppressed and parsing fails.
    """

    paper_dict = None

    if suppress_warnings:
        warnings.simplefilter("ignore")

    if suppress_errors:
        try:
            paper_dict = _actually_generate_paper_dict(pmcid, paper_root, verbose)
        except Exception as e:
            print(f"An exception occurred: {str(e)}")
    else:
        paper_dict = _actually_generate_paper_dict(pmcid, paper_root, verbose)

    if suppress_warnings:
        warnings.simplefilter("default")

    return paper_dict


def _actually_generate_paper_dict(
    pmcid: int, paper_root: ET.Element, verbose: bool = False
) -> dict:
    """
    Actual paper dictionary generation function.

    This function is called by the wrapper function `generate_paper_dict()`.
    Unlike the wrapper function, this function does not have error or warning
        suppression.

    :param int pmcid: Unique PMCID for the article being parsed.
    :param ET.Element paper_root: The root element of the PMC paper XML tree.
    :param bool verbose: Whether or not to have verbose output for debugging.

    :return: A dictionary containing relevant PMC paper XML information.
    :rtype: dict
    """
    root = paper_root
    # KEEP TRACK OF XREFS, TABLES, FIGURES IN BIMAP
    # (THIS WILL BE UPDATED DURING TEXT RETRIEVAL
    # WHEN HTML REF TAGS ARE SPLIT OUT)
    ref_map = basicBiMap()

    # STORE EXTRACTED INFO IN PAPER DICT
    paper_dict = {
        "PMCID": pmcid,
        "Title": gather_title(root),
        "Authors": gather_authors(root),
        "Non-Author Contributors": gather_non_author_contributors(root),
        "Abstract": gather_abstract(root, ref_map),
        "Body": gather_body(root, ref_map),
        "Journal ID": gather_journal_id(root),
        "Journal Title": gather_journal_title(root),
        "ISSN": gather_issn(root),
        "Publisher Name": gather_publisher_name(root),
        "Publisher Location": gather_publisher_location(root),
        "Article ID": gather_article_id(root),
        "Article Types": gather_article_types(root),
        "Article Categories": gather_article_categories(root),
        "Published Date": gather_published_date(root),
        "Volume": gather_volume(root),
        "Issue": gather_issue(root),
        "First Page": gather_fpage(root),
        "Last Page": gather_lpage(root),
        "Permissions": gather_permissions(root),
        "Funding": gather_funding(root),
        "Footnote": gather_footnote(root),
        "Acknowledgements": gather_acknowledgements(root),
        "Notes": gather_notes(root),
        "Custom Meta": gather_custom_metadata(root),
        "Ref Map With Tags": copy.deepcopy(ref_map),
        "Ref Map": _clean_ref_map(paper_root=root, ref_map=ref_map),
    }

    citations, tables, figures = _split_citations_tables_figs(paper_dict["Ref Map"])
    paper_dict["Citations"] = citations
    paper_dict["Tables"] = tables
    paper_dict["Figures"] = figures

    if verbose:
        print(
            (
                "Finished generating Paper object for "
                f"PMCID = {paper_dict['PMCID']}..."
            )
        )

    return paper_dict


def define_data_dict() -> dict:
    """
    Returns a static definition of each of the elements
    returned in a Paper dictionary.

    :return: A dictionary where keys are the elements in the Paper dictionary,
        and values are descriptions of those elements.
    :rtype: dict
    """
    data_dict = {
        "PMCID": "PMCID of the PMC article. Unique.",
        "Title": "Title of the PMC article.",
        "Authors": cleanerdoc(
            """Dataframe of the PMC Authors, including first
            names, last names, email addresses, and affiliations
            if provided."""
        ),
        "Non-Author Contributors": cleanerdoc(
            """Dataframe of the non-author contributors,
            including first names, last names, email addresses,
            and affiliations if provided."""
        ),
        "Abstract": cleanerdoc(
            """List of TextSections parsed from the abstract
            portion of the XML. Use Paper.print_abstract() or
            Paper.abstract_as_str() for a simple view of the text."""
        ),
        "Body": cleanerdoc(
            """List of TextSections parsed from the body portion of the XML.
            Use Paper.print_body() or Paper.body_as_str() for a simple view
            of the text."""
        ),
        "Journal ID": cleanerdoc(
            """Dict of ID Type, ID pairs for the Journal in which the article
            was published. ie. NLM-TA and ISO-ABBREV IDs."""
        ),
        "Journal Title": cleanerdoc("""Name of the journal in text."""),
        "ISSN": "Dict of ISSN type, ISSN number values for the article.",
        "Publisher Name": "Name of the publisher in text.",
        "Publisher Location": "Location of the publisher in text.",
        "Article ID": cleanerdoc(
            """Dict of ID Type, ID Value pairs. ie. p.article_id['pmc'] gives
            the PMCID for the article."""
        ),
        "Article Types": "List of 'header' article types for the article.",
        "Article Categories": "List of 'non-header' article types for the article.",
        "Published Date": cleanerdoc(
            """Dict of various publishing dates of the paper
            (ie: electronic pub, print pub)."""
        ),
        "Volume": cleanerdoc(
            """The Volume # in which this paper was
            published in its journal(s)."""
        ),
        "Issue": cleanerdoc(
            """The Issue # in which this paper was grouped within the
            volume of the journal(s) in which it is published."""
        ),
        "FPage": "First page on which this paper was published in its journal.",
        "LPage": "Last page on which this paper was published in its journal.",
        "Permissions": cleanerdoc(
            """Summary of copyright statement, license type,
            and full license text for the paper."""
        ),
        "Copyright Statement": cleanerdoc(
            """Returns the Copyright statement. Usually a short phrase
            identifying the individuals who have copyrighted this research,
            under a copyright license type found via paper.license."""
        ),
        "License": cleanerdoc(
            """Returns the License Type the research is licensed
             under (ie: Open Access)."""
        ),
        "Funding": cleanerdoc(
            """Returns a list of groups which funded the research.
              Important for bias detection."""
        ),
        "Footnote": "Text of any footnote statement provided with the article.",
        "Acknowledgements": cleanerdoc(
            """List of acknowledgement statements provided with
            the article."""
        ),
        "Notes": "List of notes included with the article.",
        "Custom Meta": cleanerdoc(
            """Dict of custom metadata key, value pairs
            provided with the article."""
        ),
        "Ref Map": cleanerdoc(
            """Dict of Index, Reference value pairs. Use p.ref_map to decode
            data references within TextSection.text_with_refs text. ie. When
            working with the full text with references, you may come across
            something like [MHTML::dataref::0]. This means that the reference
            under p.ref_map[0] was extracted from this location in the text.
            This can be useful for linking text with tables, figures, and
            xrefs for more detailed analysis."""
        ),
    }

    return data_dict


def gather_title(root: ET.Element) -> str:
    """
    Extract the title of a PMC paper from its XML root.

    :param ET.Element root: The root element of the PMC paper XML tree.

    :return: The title of the PMC paper.
    :rtype: str
    """
    matches = root.xpath("//article-title/text()")
    if len(matches) > 1:
        warnings.warn(
            (
                "Warning! Multiple titles matched. Setting "
                "Paper.title to the first match."
            ),
            unexpectedMultipleMatchWarning,
        )
    elif len(matches) == 0:
        warnings.warn(
            "No article title found in the retrieved XML.", unexpectedZeroMatchWarning
        )
        return None
    title = matches[0]

    return title


def _get_contributor_tuples(
    root: ET.Element, contributors: List[ET.Element]
) -> List[Tuple]:
    """
    Helper function to retrieve tuples of contributor information.

    :param ET.Element root: The root of the XML tree to search.
    :param List[ET.Element] contributors: A list of lxml Element objects
        containing contributor information.

    :return: A list of tuples representing contributor information in the form
        (contrib_type, first_name, last_name, address, affiliations).
    :rtype: List[Tuple]
    """
    contributor_tuples = []
    for contributor in contributors:
        contrib_type = contributor.get("contrib-type").capitalize()
        if contrib_type:
            contrib_type = contrib_type.strip()
        first_name = contributor.findtext(".//given-names")
        if first_name is not None:
            first_name = first_name.strip()
        last_name = contributor.findtext(".//surname")
        if last_name is not None:
            last_name = last_name.strip()
        address = contributor.findtext(".//address/email")
        if address is not None:
            address = address.strip()
        affiliations = []
        aff_paths = contributor.xpath(".//xref[@ref-type='aff']")
        for aff in aff_paths:
            aff_id = aff.get("rid")
            aff_texts = root.xpath(
                (f"//contrib-group/aff[@id='{aff_id}']" "/text()[not(parent::label)]")
            )
            if len(aff_texts) > 1:
                warnings.warn(
                    (
                        "Multiple affiliations with the same ID found. "
                        "Check XML Formatting."
                    ),
                    unexpectedMultipleMatchWarning,
                )
            if len(aff_texts) == 0:
                aff_texts = ["Affiliation data not found."]

            institutions = root.xpath(
                (
                    f"//contrib-group/aff[@id='{aff_id}']"
                    "/institution-wrap/institution/text()"
                )
            )
            institutions = " ".join([str(inst) for inst in institutions])

            # Generate affiliation text
            affiliation = ""
            if institutions:
                affiliation = f"{aff_id.strip()}: {institutions}{aff_texts[0].strip()}"
            else:
                affiliation = f"{aff_id.strip()}: {aff_texts[0].strip()}"
            affiliations.append(affiliation)

        contributor_tuples.append(
            (contrib_type, first_name, last_name, address, affiliations)
        )
    return contributor_tuples


def gather_authors(root: ET.Element) -> pd.DataFrame:
    """
    Extract authors, their emails, and affiliations from a PMC XML.

    :param ET.Element root: The root element of the PMC paper XML tree.

    :return: A DataFrame containing author information with columns:
             - Contributor_Type: Type of contributor (e.g., 'author').
             - First_Name: First name of the author.
             - Last_Name: Last name of the author.
             - Email_Address: Email address of the author.
             - Affiliations: Affiliations of the author.
    :rtype: pd.DataFrame
    """
    authors = root.xpath(".//contrib[@contrib-type='author']")
    if len(authors) == 0:
        warnings.warn(
            "Warning! Authors could not be matched", unexpectedZeroMatchWarning
        )
        return None

    # Extract the first and last names of the authors and store them in a list
    author_tuples = _get_contributor_tuples(root=root, contributors=authors)

    authors_df = pd.DataFrame(author_tuples)
    authors_df.columns = [
        "Contributor_Type",
        "First_Name",
        "Last_Name",
        "Email_Address",
        "Affiliations",
    ]

    return authors_df


def gather_non_author_contributors(root: ET.Element) -> Union[str, pd.DataFrame]:
    """
    Extract non-author contributors from a PMC XML.

    :param ET.Element root: The root element of the PMC paper XML tree.

    :return: Either a string indicating that no non-author contributors were found,
             or a DataFrame containing contributor information with columns:
             - Contributor_Type: Type of contributor.
             - First_Name: First name of the contributor.
             - Last_Name: Last name of the contributor.
             - Email_Address: Email address of the contributor.
             - Affiliations: Affiliations of the contributor.
    :rtype: Union[str, pd.DataFrame]
    """

    return_val = "No non-author contributors were found after parsing this paper."

    non_author_contributors = root.xpath(".//contrib[not(@contrib-type='author')]")
    if len(non_author_contributors) > 0:
        non_author_tuples = _get_contributor_tuples(
            root=root, contributors=non_author_contributors
        )
        non_authors_df = pd.DataFrame(non_author_tuples)
        non_authors_df.columns = [
            "Contributor_Type",
            "First_Name",
            "Last_Name",
            "Email_Address",
            "Affiliations",
        ]
        return_val = non_authors_df

    return return_val


def gather_abstract(
    root: ET.Element, ref_map: basicBiMap
) -> List[Union[TextSection, TextParagraph]]:
    """
    Extract all abstract text sections from an XML document and return them as
    a list of TextSections and/or TextParagraphs.

    :param ET.Element root: The root element of the PMC paper XML tree.
    :param basicBiMap ref_map: A reference map used for decoding data references
        within the text.

    :return: A list of TextSections and/or TextParagraphs representing the abstract
             text sections in the XML.
    :rtype: List[Union[TextSection, TextParagraph]]
    """
    abstract = []

    # get abstract subtree from XML
    matches = root.xpath("//abstract")
    if len(matches) > 1:
        warnings.warn(
            (
                "Warning! Multiple abstracts matched. Filling in Paper.abstract "
                "with the first match."
            ),
            unexpectedMultipleMatchWarning,
        )
    elif len(matches) == 0:
        warnings.warn("No abstract found.", unexpectedZeroMatchWarning)
        return None
    abstract_root = matches[0]

    # iterate through abstract subtree and add in text sections (recursive)
    # and text paragraphs (flat)
    for child in abstract_root.iterchildren():
        if child.tag == "sec":
            abstract.append(TextSection(sec_root=child, ref_map=ref_map))
        elif child.tag == "p":
            abstract.append(TextParagraph(p_root=child, ref_map=ref_map))
        else:
            warnings.warn(
                (
                    f"Warning! Unexpected child with of type {child.tag} found "
                    "under an XML <abstract> tag."
                )
            )

    return abstract


def gather_body(
    root: ET.Element, ref_map: basicBiMap
) -> List[Union[TextSection, TextParagraph]]:
    """
    Extract all body text sections from an XML document and return them as
    a list of TextSections and/or TextParagraphs.

    :param ET.Element root: The root element of the PMC paper XML tree.
    :param basicBiMap ref_map: A reference map used for decoding data references
        within the text.

    :return: A list of TextSections and/or TextParagraphs representing the body
             text sections in the XML.
    :rtype: List[Union[TextSection, TextParagraph]]
    """
    body = []

    # get abstract subtree from XML
    matches = root.xpath("//body")
    if len(matches) > 1:
        warnings.warn(
            (
                "Warning! Multiple 'body's matched. "
                "Filling in Paper.body with the first match."
            ),
            unexpectedMultipleMatchWarning,
        )
    elif len(matches) == 0:
        warnings.warn(
            (
                "Warning! No <body> tag found. This paper may be abstract only, "
                "or the Open Access portion may be abstract only. This also may "
                "happen with author manuscripts and other non-final editions."
            )
        )
        return None
    body_root = matches[0]

    # iterate through body subtree and add in text sections (recursive)
    # and text paragraphs (flat)
    for child in body_root.iterchildren():
        if child.tag == "sec":
            body.append(TextSection(sec_root=child, ref_map=ref_map))
        elif child.tag == "p":
            body.append(TextParagraph(p_root=child, ref_map=ref_map))
        else:
            warnings.warn(
                (
                    f"Warning! Unexpected child with of type {child.tag} found "
                    "under an XML <body> tag."
                )
            )

    return body


def gather_journal_id(root: ET.Element) -> dict:
    """
    Extract Journal IDs from a PMC XML document.

    :param ET.Element root: The root element of the PMC paper XML tree.

    :return: A dictionary containing Journal IDs with the ID type as keys and
             corresponding values as the ID values.
    :rtype: dict
    """
    journal_ids = root.xpath("//journal-meta/journal-id")
    id_dict = {
        journal_id.get("journal-id-type"): journal_id.text for journal_id in journal_ids
    }

    return id_dict


def gather_journal_title(root: ET.Element) -> Union[List[str], str]:
    """
    Extract Journal Title(s) from a PMC XML document.

    :param ET.Element root: The root element of the PMC paper XML tree.

    :return: Either a string representing the Journal Title if there's only one,
             a list of strings representing multiple Journal Titles if there are
             multiple, or None if no journal title is found.
    :rtype: Union[List[str], str, None]
    """
    return_val = None
    titles = []
    title_matches = root.xpath("//journal-title")
    for title in title_matches:
        titles.append(title.text)
    # might have multiple journals & journal titles
    if len(titles) > 1:
        return_val = titles
    elif len(titles) == 0:
        warnings.warn("No journal title found.", unexpectedZeroMatchWarning)
        return_val = None
    else:
        return_val = titles[0]
    return return_val


def gather_issn(root: ET.Element) -> dict:
    """
    Extract ISSN values from a PMC XML document.

    :param ET.Element root: The root element of the PMC paper XML tree.

    :return: A dictionary containing ISSN values with the publication type
             as keys and corresponding values as the ISSN numbers.
    :rtype: dict
    """
    issns = root.xpath("//journal-meta/issn")
    issn_dict = {issn.get("pub-type"): issn.text for issn in issns}

    return issn_dict


def gather_publisher_name(root: ET.Element) -> Union[str, List[str]]:
    """
    Extract Publisher Name(s) from a PMC XML document.

    :param ET.Element root: The root element of the PMC paper XML tree.

    :return: Either a string representing the Publisher Name if there's only one,
             or a list of strings representing multiple Publisher Names if there are
             multiple.
    :rtype: Union[str, List[str]]
    """
    publisher_name_or_names = None
    publishers = root.xpath("//journal-meta/publisher/publisher-name")
    if len(publishers) == 1:
        publisher_name_or_names = publishers[0].text
    else:
        publisher_name_or_names = [publisher.text for publisher in publishers]
    return publisher_name_or_names


def gather_publisher_location(root: ET.Element) -> Union[str, List[str]]:
    """
    Extract Publisher Location(s) from a PMC XML document.

    :param ET.Element root: The root element of the PMC paper XML tree.

    :return: Either a string representing the Publisher Location if there's
            only one, or a list of strings representing multiple Publisher
            Locations if there are multiple.
    :rtype: Union[str, List[str]]
    """
    publisher_loc_or_locs = None
    publisher_locs = root.xpath("//journal-meta/publisher/publisher-loc")
    if len(publisher_locs) == 1:
        publisher_loc_or_locs = publisher_locs[0].text
    else:
        publisher_loc_or_locs = [publisher_loc.text for publisher_loc in publisher_locs]
    return publisher_loc_or_locs


def gather_article_id(root: ET.Element) -> Dict[str, str]:
    """
    Gather Article IDs from PMC XML.
    """
    article_ids = root.xpath("//article-meta/article-id")
    id_dict = {
        article_id.get("pub-id-type"): article_id.text for article_id in article_ids
    }

    return id_dict


def gather_article_types(root: ET.Element) -> List[str]:
    """
    Extract Article Types from a PMC XML document.

    Article Types are article-categories marked by the subj-group-type 'heading'.

    :param ET.Element root: The root element of the PMC paper XML tree.

    :return: A list of strings representing the Article Types found in the XML.
    :rtype: List[str]
    """
    matches = root.xpath("//article-meta/article-categories")
    if len(matches) > 1:
        warnings.warn(
            (
                "Warning! Multiple 'article-categories' lists matched. "
                "Filling in Paper.article_categories with the first match."
            ),
            unexpectedMultipleMatchWarning,
        )
    elif len(matches) == 0:
        warnings.warn("No 'article-categories' list found.", unexpectedZeroMatchWarning)
        return None

    article_categories = matches[0]
    heading_categories = article_categories.xpath(
        "subj-group[@subj-group-type='heading']/subject"
    )
    heading_cats = [heading_cat.text for heading_cat in heading_categories]

    if not heading_cats:
        heading_cats = (
            "No article type found (No article category with "
            "subject type 'heading'). Check "
            "Paper.article_categories for other categories."
        )
    return heading_cats


def gather_article_categories(root: ET.Element) -> List[str]:
    """
    Extract Other Article Categories from a PMC XML document.

    This function retrieves article categories that are not marked as 'heading'
    in the subj-group-type attribute.

    :param ET.Element root: The root element of the PMC paper XML tree.

    :return: A list of dictionaries containing other article categories with
             the subj-group-type as keys and corresponding category values as
             values.
    :rtype: List[Dict[str, str]]
    """
    matches = root.xpath("//article-meta/article-categories")
    if len(matches) > 1:
        warnings.warn(
            (
                "Warning! Multiple 'article-categories' lists matched. Filling "
                "in Paper.article_categories with the first match."
            ),
            unexpectedMultipleMatchWarning,
        )
    elif len(matches) == 0:
        warnings.warn("No 'article-categories' list found.", unexpectedZeroMatchWarning)
        return None
    article_categories = matches[0]
    other_categories = article_categories.xpath(
        "subj-group[not(@subj-group-type='heading')]/subject"
    )
    other_cats = [
        {other_cat.get("subj-group-type"): other_cat.text}
        for other_cat in other_categories
    ]

    if not other_cats:
        other_cats = (
            "No extra article categories found. "
            "Check paper.article_types for header categories."
        )
    return other_cats


def gather_published_date(root: ET.Element) -> Dict[str, datetime]:
    """
    Extract Publishing Dates from a PMC XML document.

    This function gathers electronic publishing, print publishing, and other
    dates from the article metadata.

    :param ET.Element root: The root element of the PMC paper XML tree.

    :return: A dictionary containing publishing dates with the publication type
             as keys and corresponding datetime values as values.
    :rtype: Dict[str, datetime]
    """
    # TODO: update for multi-publishing (need to find an example first)

    pdate_dict = {}
    matches = root.xpath("//article-meta/pub-date")
    for match in matches:
        pub_type = match.get("pub-type")

        year = 1  # default
        year_matches = match.xpath("year/text()")
        if len(year_matches) > 0:
            year = int(year_matches[0])
        else:
            warnings.warn(
                (
                    "No year found for one of the publishing dates. "
                    "Defaulting to year = 1!"
                ),
                unexpectedZeroMatchWarning,
            )

        # if not month found, assume the 1st (standard practice - )
        month = 1
        month_matches = match.xpath("month/text()")
        if len(month_matches) > 0:
            month = int(month_matches[0])

        # if no day found, assume the 1st (standard practice)
        day = 1
        day_matches = match.xpath("day/text()")
        if len(day_matches) > 0:
            day = int(day_matches[0])

        full_date = datetime(year=year, month=month, day=day)
        pdate_dict[pub_type] = full_date
    return pdate_dict


def gather_volume(root: ET.Element) -> str:
    """
    Extract the Volume # of the Parent Publication from a PMC XML document.

    :param ET.Element root: The root element of the PMC paper XML tree.

    :return: A string representing the Volume # of the parent publication,
             or None if no Volume # is found.
    :rtype: str or None
    """
    # TODO: update for multi-publishing (need to find an example first)

    matches = root.xpath("//article-meta/volume/text()")
    volume = None
    if len(matches) == 0:
        warnings.warn("No Volume # found for Publication.", unexpectedZeroMatchWarning)
    else:
        volume = matches[0]

    return volume


def gather_issue(root: ET.Element) -> str:
    """
    Extract the Issue # of the Parent Publication from a PMC XML document.

    :param ET.Element root: The root element of the PMC paper XML tree.

    :return: A string representing the Issue # of the parent publication,
             or None if no Issue # is found.
    :rtype: str or None
    """
    # TODO: update for multi-publishing (need to find an example first)

    matches = root.xpath("//article-meta/issue/text()")
    issue = None
    if len(matches) == 0:
        warnings.warn("No Issue # found for Publication.", unexpectedZeroMatchWarning)
    else:
        issue = matches[0]

    return issue


def gather_fpage(root: ET.Element) -> str:
    """
    Extract the First Page Number of this article in its parent publication
    from a PMC XML document.

    :param ET.Element root: The root element of the PMC paper XML tree.

    :return: A string representing the First Page Number of the article
             in its parent publication, or None if no First Page # is found.
    :rtype: str or None
    """
    # TODO: update for multi-publishing (need to find an example first)

    matches = root.xpath("//article-meta/fpage/text()")
    fpage = None
    if len(matches) == 0:
        warnings.warn(
            "No First Page # found for Publication.", unexpectedZeroMatchWarning
        )
    else:
        fpage = matches[0]

    return fpage


def gather_lpage(root: ET.Element) -> str:
    """
    Extract the Last Page Number of this article in its parent publication
    from a PMC XML document.

    :param ET.Element root: The root element of the PMC paper XML tree.

    :return: A string representing the Last Page Number of the article
             in its parent publication, or None if no Last Page # is found.
    :rtype: str or None
    """
    # TODO: update for multi-publishing (need to find an example first)

    matches = root.xpath("//article-meta/lpage/text()")
    lpage = None
    if len(matches) == 0:
        warnings.warn(
            "No Last Page # found for Publication.", unexpectedZeroMatchWarning
        )
    else:
        lpage = matches[0]

    return lpage


def gather_permissions(root: ET.Element) -> Dict[str, str]:
    """
    Extract permissions information from a PMC XML document.

    This function retrieves the copyright statement, license type, and license
    text from the article metadata.

    :param ET.Element root: The root element of the PMC paper XML tree.

    :return: A dictionary containing the following keys:
             - "Copyright Statement": A string representing the copyright statement.
             - "License Type": A string representing the license type.
             - "License Text": A string containing the license text.
    :rtype: Dict[str, str]
    """
    copyright_statement_matches = root.xpath(
        "//article-meta/permissions/copyright-statement/text()"
    )
    copyright_statement = "No copyright statement found."
    if len(copyright_statement_matches) == 0:
        warnings.warn("No copyright statement found.", unexpectedZeroMatchWarning)
    elif len(copyright_statement_matches) > 1:
        warnings.warn(
            ("Multiple copyright statements found. " "Retrieving the first statement."),
            unexpectedMultipleMatchWarning,
        )
    else:
        copyright_statement = copyright_statement_matches[0]

    license_matches = root.xpath("//article-meta/permissions/license")
    if len(license_matches) == 0:
        warnings.warn("No license found.", unexpectedZeroMatchWarning)
        return None
    elif len(license_matches) > 1:
        warnings.warn(
            "Multiple licenses found. Retrieving the first statement.",
            unexpectedMultipleMatchWarning,
        )
    license = license_matches[0]
    license_type = license.get("license-type")
    if not license_type:
        license_type = "Not Specified"
    license_text = []
    for child in license.iterchildren():
        if child.tag == "license-p":
            license_text.append(TextParagraph(p_root=child))
        else:
            warnings.warn(
                (
                    f"Warning! Unexpected child with of type {child.tag} found "
                    "under an XML <license> tag."
                )
            )
    license_text = "\n".join([str(par) for par in license_text])

    permissions_dict = {
        "Copyright Statement": copyright_statement,
        "License Type": license_type,
        "License Text": license_text,
    }
    return permissions_dict


def gather_funding(root: ET.Element) -> List[str]:
    """
    Extract funding information from a PMC XML document.

    This function retrieves a list of funding institutions mentioned in the
    article metadata.

    :param ET.Element root: The root element of the PMC paper XML tree.

    :return: A list of strings representing funding institutions, or None if
             no funding information is found.
    :rtype: List[str] or None
    """
    matches = root.xpath("//article-meta/funding-group")
    funding_institutions = []
    for match in matches:
        institutions = match.xpath("award-group/funding-source/institution/text()")
        funding_institutions.extend([inst for inst in institutions])

    if len(funding_institutions) == 0:
        funding_institutions = None
    return funding_institutions


def gather_footnote(root: ET.Element) -> str:
    """
    Extract footnote information from a PMC XML document.

    This function retrieves and concatenates footnotes found in the article's
    back matter.

    :param ET.Element root: The root element of the PMC paper XML tree.

    :return: A string containing the concatenated footnotes, or None if no
             footnotes are found.
    :rtype: str or None
    """
    matches = root.xpath("//back/fn-group/fn")
    footnote = ""
    for fn in matches:
        for child in fn:
            if child.tag == "p":
                if len(footnote) == 0:
                    footnote += str(TextParagraph(p_root=child))
                else:
                    footnote += " - " + str(TextParagraph(p_root=child))
            else:
                warnings.warn(
                    (
                        f"Unexpected child of type {child.tag} under a footnote "
                        "(<fn>) tag. Ignoring."
                    )
                )

    if len(footnote) == 0:
        footnote = None

    return footnote


def gather_acknowledgements(root: ET.Element) -> Union[List[str], str]:
    """
    Extract acknowledgements information from a PMC XML document.

    This function retrieves a list of acknowledgements found in the article's
    XML tree.

    :param ET.Element root: The root element of the PMC paper XML tree.

    :return: A list of strings representing acknowledgements, or a string
             indicating that no acknowledgements were found.
    :rtype: Union[List[str], str]
    """
    matches = root.xpath("//ack")
    acknowledgements = [" ".join(match.itertext()).strip() for match in matches]

    return acknowledgements


def gather_notes(root: ET.Element) -> str:
    """
    Extract notes information from a PMC XML document.

    This function retrieves a list of notes found in the article's XML tree.

    :param ET.Element root: The root element of the PMC paper XML tree.

    :return: A list of strings representing notes, or an empty list if no notes
             are found.
    :rtype: List[str]
    """
    notes = []
    matches = root.xpath("//notes")
    notes = [
        stringify_note(note) for note in matches if not note.getparent().tag == "notes"
    ]

    return notes


def stringify_note(root: ET.Element) -> str:
    """
    Recursively convert a notes section into a string.

    This function traverses the XML tree of a notes section and recursively
    converts it into a string. It includes the <title>, <p>, and child <notes>
    content.

    :param ET.Element root: The root element of the notes section in the PMC
                            paper XML tree.

    :return: A string representation of the notes section.
    :rtype: str
    """
    note = ""
    for child in root.iterchildren():
        if child.tag == "title":
            note += f"Title: {child.text}\n"
        elif child.tag == "p":
            note += child.text
        elif child.tag == "notes":
            note += "\n" + textwrap.indent(stringify_note(child), " " * 4)
    note += "\n"
    return note


# def _get_note(note_root: ET.Element) ->


def gather_custom_metadata(root: ET.Element) -> Dict[str, str]:
    """
    Extract custom metadata key-value pairs from a PMC XML document.

    This function retrieves custom metadata key-value pairs found in the
    article's XML tree. Custom metadata consists of user-defined key-value pairs.

    :param ET.Element root: The root element of the PMC paper XML tree.

    :return: A dictionary containing custom metadata key-value pairs, or None if
             no custom metadata is found.
    :rtype: Dict[str, str] or None
    """
    custom = {}
    matches = root.xpath("//custom-meta")
    for custom_meta in matches:
        meta_name = custom_meta.find("meta-name")
        if meta_name is not None:
            meta_name = meta_name.text
        meta_value = custom_meta.find("meta-value")
        meta_data = None
        if meta_value is not None:
            meta_data = " ".join(meta_value.itertext())
        if meta_data:
            if meta_name is None:
                meta_name = (
                    uuid.uuid4()
                )  # give random unique identifier if no meta key found
            custom[meta_name] = meta_data

    if len(custom) == 0:
        custom = None
    return custom


def _parse_citation(
    citation_root: ET.Element,
) -> Union[Dict[str, Union[List[str], str]], str]:
    """
    Parse citation information from a citation XML element.

    This function parses citation information from a given XML element
    representing a citation. It attempts to extract the following information:

    - Authors: List of author names.
    - Title: Title of the cited work.
    - Source: Source of the cited work (e.g., journal name).
    - Year: Year of publication.
    - Volume: Volume number (if applicable).
    - FirstPage: First page number (if applicable).
    - LastPage: Last page number (if applicable).
    - DOI: DOI (Digital Object Identifier) of the cited work (if available).
    - PMID: PubMed Identifier of the cited work (if available).

    If the function successfully extracts information, it returns a dictionary
    containing the parsed data. If no information can be extracted, it returns
    either an empty dictionary or the raw citation text (if available).

    :param ET.Element citation_root: The root element of the citation in the XML.

    :return: A dictionary containing parsed citation information or the raw
        citation text.
    :rtype: Union[Dict[str, Union[List[str], str]], str]
    """

    root = citation_root

    # Find authors in common element-citation format
    author_matches = root.xpath('.//person-group[@person-group-type="author"]/name')

    # If failed, try to find full citation in mixed-citation format-----------
    mixed_citation = None
    if len(author_matches) == 0:
        mixed_citation = root.xpath("//mixed-citation/text()")
        if len(mixed_citation) > 0:
            return str(mixed_citation[0])
    # ------------------------------------------------------------------------

    # If still failed, raise a warning.
    if len(author_matches) == 0:
        warnings.warn(
            f"No authors found in citation {root.get('id')}", unexpectedZeroMatchWarning
        )

    # tries to retrieve all of the following info, fails silently
    # if none found since many refs incomplete
    citation_dict = {
        "Authors": [
            (
                f"{_try_get_xpath_text(name, 'given-names')} "
                f"{_try_get_xpath_text(name, 'surname')}"
            )
            for name in author_matches
            if author_matches
        ],
        "Title": _try_get_xpath_text(root, ".//article-title"),
        "Source": _try_get_xpath_text(root, ".//source"),
        "Year": _try_get_xpath_text(root, ".//year"),
        "Volume": _try_get_xpath_text(root, ".//volume"),
        "FirstPage": _try_get_xpath_text(root, ".//fpage"),
        "LastPage": _try_get_xpath_text(root, ".//lpage"),
        "DOI": _try_get_xpath_text(root, './/pub-id[@pub-id-type="doi"]'),
        "PMID": _try_get_xpath_text(root, './/pub-id[@pub-id-type="pmid"]'),
    }

    return citation_dict


def _try_get_xpath_text(root: ET.Element, xpath: str, verbose=False) -> str:
    """
    Attempt to retrieve the text content of an XML element using an XPath expression.

    This function tries to find the first element matching the given XPath expression
    within the specified XML element and retrieves its text content. If successful, it
    returns the text content; otherwise, it returns None.

    :param ET.Element root: The XML element from which to retrieve the text.
    :param str xpath: The XPath expression used to locate the target element.
    :param bool verbose: If True, warnings will be issued for failed retrieval attempts.

    :return: The text content of the matching XML element or None if not found.
    :rtype: str
    """
    return_text = None
    try:
        return_text = root.find(xpath).text
    except AttributeError:
        if verbose:
            warnings.warn(
                (
                    "Failed xpath text retrieval while trying to find "
                    f"{xpath}.text(). Root ID: {root.get('id')}"
                )
            )

    return return_text


def _find_key_of_xpath(ref_map: basicBiMap, xpath_query: str) -> int:
    """
    Search for the first key in the reference map (ref_map) where the value
    matches the provided XPath query.

    This function iterates through the reference map, which associates keys with
    XPath expressions (values), and checks if the specified XPath query matches the
    value associated with each key. The first matching key found is returned.

    :param basicBiMap ref_map: A bidirectional map containing XPath expressions as
        values and associated keys.
    :param str xpath_query: The XPath query to match against the values in the ref_map.

    :return: The first key in the ref_map where the XPath query matches the value,
        or None if no match is found.
    :rtype: int or None
    """

    ref_map = copy.deepcopy(ref_map)
    # Iterate through the dictionary and find the key with matching value
    matching_key = None
    for key, value in ref_map.items():
        if len(ET.fromstring(value).xpath(xpath_query)) > 0:
            matching_key = key
            break

    return matching_key


def _clean_ref_map(paper_root: ET.Element, ref_map: basicBiMap) -> basicBiMap:
    """
    Process a reference map (ref_map) by replacing various types of references with
    their corresponding information, such as citations, tables, and figures.

    This function iterates through the reference map and processes different types of
    references found in the map. It replaces bibliography (bibr) references with actual
    citation information, table references with table data, and figure references with
    figure information. The resulting cleaned reference map contains the processed
    references.

    :param ET.Element paper_root: The root element of the paper's XML.
    :param basicBiMap ref_map: A bidirectional map containing keys and associated
        values that represent different types of references.

    :return: A cleaned reference map with references replaced by their respective
        information.
    :rtype: basicBiMap
    """
    cleaned_ref_map = {}

    for key, item in ref_map.items():
        root = ET.fromstring(item)

        # -------XREFS LINK TO ACTUAL ITEMS OR FILL WITH BIBR--------------
        # process xrefs to citations, tables, and figures
        if root.tag == "xref":
            if root.get("ref-type") == "bibr":
                ref_id = root.get("rid")
                if not ref_id:
                    warnings.warn(
                        (
                            "Citation without a reference id specified "
                            f"(Citation {root.text})!"
                        ),
                        unmatchedCitationWarning,
                    )
                    continue

                # XPath expression to find the <ref>
                # element based on the reference ID
                matching_citation_expr = f"//ref[@id='{ref_id}']"
                matches = paper_root.xpath(matching_citation_expr)
                if len(matches) == 0:
                    warnings.warn(
                        (
                            "Citation without matching reference "
                            f"(Citation {root.text})!"
                        ),
                        unmatchedCitationWarning,
                    )
                    continue
                elif len(matches) > 1:
                    warnings.warn(
                        (
                            "Multiple references found for a single citation. "
                            "Filling in with the first match."
                        )
                    )

                reference_xml = matches[0]
                cleaned_reference = _parse_citation(reference_xml)
                cleaned_ref_map[key] = cleaned_reference

            elif root.get("ref-type") == "table":
                table_id = root.get("rid")
                if not table_id:
                    warnings.warn(
                        """Table ref without reference ID, no table will
                        be matched!""",
                        unmatchedTableWarning,
                    )
                    continue

                table_xpath = f"//table-wrap[@id='{table_id}']"
                matches = paper_root.xpath(table_xpath)
                if len(matches) == 0:
                    warnings.warn(
                        (f"Table xref with rid={table_id} not " "matched in the XML!"),
                        unmatchedTableWarning,
                    )
                    continue
                elif len(matches) > 1:
                    warnings.warn(
                        (
                            "Multiple references found for a single table. "
                            "Filling in with the first match."
                        )
                    )
                table_root = matches[0]
                cleaned_ref_map[key] = TextTable(table_root=table_root)

            elif root.get("ref-type") == "fig":
                fig_id = root.get("rid")
                if not fig_id:
                    warnings.warn(
                        (
                            "Figure ref unmatched. Figure ref without matching "
                            f"figure (Figure {root.text})!"
                        ),
                        unmatchedFigureWarning,
                    )
                    continue

                fig_xpath = f"//fig[@id='{fig_id}']"
                matches = paper_root.xpath(fig_xpath)
                if len(matches) == 0:
                    warnings.warn(
                        (f"Figure xref with rid={fig_id} not matched " "in the XML!"),
                        unmatchedFigureWarning,
                    )
                    continue
                elif len(matches) > 1:
                    warnings.warn(
                        (
                            "Multiple references found for a single figure. "
                            "Filling in with the first match."
                        )
                    )
                fig_root = matches[0]
                cleaned_ref_map[key] = TextFigure(fig_root=fig_root)

            elif root.get("ref-type"):
                warnings.warn(
                    (
                        f"Unknown reference type: {root.get('ref_type')} "
                        "found in ref_map."
                    )
                )
            else:
                warnings.warn(
                    (
                        "<xref> in ref_map with no ref-type specified. "
                        f"Ignoring. ({root.text})"
                    )
                )

        # process tables that are directly in the ref map
        elif root.tag == "table-wrap":
            cleaned_ref_map[key] = TextTable(table_root=root)
        # process figures that are directly in the ref map
        elif root.tag == "fig":
            cleaned_ref_map[key] = TextFigure(fig_root=root)
        else:
            warnings.warn(
                (
                    f"Unexpected tag of type {root.tag} found in ref map. "
                    "Leaving as is instead of cleaning."
                )
            )
            cleaned_ref_map[key] = ET.tostring(root)

    # Final pass to set up links now that everything should be filled in
    for key, item in cleaned_ref_map.items():
        if isinstance(item, int):
            link_index = item
            cleaned_ref_map[key] = cleaned_ref_map[link_index]

    return cleaned_ref_map


def _get_ref_type(value):
    """
    Determine the type of reference (table, citation, or figure) based on the value
    in the reference map.

    This function examines the value to identify the type of reference it represents.
    It checks if the value is a dictionary with a "Caption" key (indicating a figure),
    a dictionary with "Authors" key (indicating a citation), an instance of TextFigure
    (indicating a figure), or an instance of TextTable (indicating a table).

    :param value: The value representing a reference in the reference map.

    :return: A string indicating the type of reference (table, citation, or figure),
        or None if the type cannot be determined.
    :rtype: str or None
    """
    ref_type = None
    if isinstance(value, dict):
        if "Caption" in value:
            ref_type = "fig"
        elif "Authors" in value:
            ref_type = "citation"
    elif isinstance(value, str):
        # if string, probably a citation scraped via
        # the mixed citation element parsing
        ref_type = "citation"
    elif isinstance(value, TextFigure):
        ref_type = "fig"
    elif isinstance(value, TextTable):
        ref_type = "table"

    return ref_type


def _get_unique_tables(table_list: List[pd.DataFrame]) -> List[dict]:
    """
    TODO: Given a set of tables (pd.DataFrame or Stylers), return a unique set.
    """
    return table_list


def _get_unique_citations(citation_list: List[dict]) -> List[dict]:
    """
    TODO: Given a set of citations, return a unique set
    based on citation['PMID']
    """
    return citation_list


def _get_unique_figures(fig_list: List[dict]) -> List[dict]:
    """
    TODO: Given a set of figures, return a unique set of figures
    based on fig['Label']
    """
    return fig_list


def _split_citations_tables_figs(
    ref_map: basicBiMap,
) -> Tuple[
    Set[Union[Dict[str, Union[List[str], str]], str]],
    Set[pd.DataFrame],
    Set[Dict[str, str]],
]:
    """
    Split the reference map into three separate sets: citations, tables, and figures.

    This function iterates through the reference map and categorizes each reference
    based on its type (citation, table, or figure). It returns three sets containing
    citations (as dictionaries or strings), tables (as pandas DataFrames),
    and figures (as dictionaries).

    :param ref_map: The reference map to be split into citations, tables, and figures.
    :type ref_map: basicBiMap

    :return: A tuple containing three sets:
        - The first set contains citations, each represented as a dictionary or string.
        - The second set contains tables, each as a pandas DataFrame.
        - The third set contains figures, each represented as a dictionary.
    :rtype: tuple
    """
    citations = []
    tables = []
    figures = []
    for i, ref in ref_map.items():
        if _get_ref_type(ref) == "citation":
            citations.append(ref)
        elif _get_ref_type(ref) == "table":
            tables.append(ref.df)
        elif _get_ref_type(ref) == "fig":
            figures.append(ref.fig_dict)
        else:
            warnings.warn(
                f"Issue finding Reference type for index {i} in reference map."
            )

    return (
        _get_unique_citations(citations),
        _get_unique_tables(tables),
        _get_unique_figures(figures),
    )


# ----------------END GENERATE PAPER DICTIONARY GIVEN PMCID----------------
